# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
from __future__ import annotations

import uuid
from collections.abc import Iterable
from dataclasses import dataclass, field, fields
from typing import List, Optional

from ._provenance import SystemProvenance

# pylint: disable=too-many-instance-attributes


@dataclass
class System:
    UUID: str = field(default_factory=lambda: str(uuid.uuid4()))
    captureStart: Optional[int] = None
    captureEnd: Optional[int] = None
    name: Optional[str] = None
    officialName: Optional[str] = None
    vendor: Optional[List[str]] = None
    description: Optional[str] = None
    provenance: Optional[List[SystemProvenance]] = None

    def merge(self, sy: System):
        if sy and self != sy:
            # leave UUID and captureTime the same
            single_value_fields = [
                "captureStart",
                "captureEnd",
                "name",
                "officialName",
                "description",
            ]
            array_fields = ["vendor", "provenance"]
            for fld in fields(self):
                if fld.name in single_value_fields:
                    current_value = getattr(self, fld.name)
                    new_value = getattr(sy, fld.name)
                    update = None
                    # If field is captureStart or captureEnd, replace if increasing time bounds
                    if fld.name == "captureStart":
                        update = new_value if new_value < current_value else current_value
                    elif fld.name == "captureEnd":
                        update = new_value if new_value > current_value else current_value
                    else:
                        update = new_value if new_value != current_value else current_value
                    self._update_field(fld.name, update)
                # for lists, append new values that we don't currently have
                if fld.name in array_fields:
                    current_arr = getattr(self, fld.name)
                    new_arr = getattr(sy, fld.name)
                    # if the multi-value fields differ, the one with new values *must* be Iterable
                    if current_arr != new_arr and isinstance(new_arr, Iterable):
                        # if our field is not iterable, initialize it as a new list
                        if current_arr is None:
                            setattr(self, fld.name, [])
                            current_arr = getattr(self, fld.name)
                        for new_value in new_arr:
                            # special case, UUID in containerPaths need updating to match our UUID
                            if new_value not in current_arr:
                                current_arr.append(new_value)
        return self.UUID, sy.UUID



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import pathlib
from collections.abc import Iterable
from typing import List, Optional

import surfactant.plugin
from surfactant.sbomtypes import SBOM, Relationship, Software

from ._internal.posix_utils import posix_normpath


def has_required_fields(metadata) -> bool:
    # no elfDependencies info, can't establish relationships
    return "elfDependencies" in metadata


# Information on loading dynamic libraries from ld.so man page:
# 1. if dependency string contains a slash, it is interpreted as a (relative or absolute) pathname and shared object is loaded using that
# 2. if no slash, search in the following order:
# - use dirs in DT_RPATH if present and DT_RUNPATH doesn't exist (deprecated)
# - use LD_LIBRARY_PATH unless being run in secure-execution mode (in which case ignored)
# - use dirs in DT_RUNPATH if present; only searched to find objects required by DT_NEEDED (direct dependencies) and don't apply to objects' children which must have their own DT_RUNPATH entries (this differs from DT_RPATH, which applies to searches for all children in the dependency tree)
# - from /etc/ld.so.cache, unless binary linked with -z nodeflib linker option then shared objects in the default paths are skipped (shared objects in hardware capability dirs are preferred over other shared objects)
# - in default path /lib, and then /usr/lib (/lib64 and /usr/lib64 maybe for 64-bit libraries); skipped if binary was linked with -z nodeflib linker option
# some strings are expanded; $ORIGIN, $LIB, $PLATFORM
# /etc/ld.so.conf can be used to configure the dynamic loader to search for other directories (such as /usr/local/lib or /opt/lib) as well; format is separate lines naming additional directories, and include <path> statement that can use wildcards to include paths from additional files, such as /etc/ld.so.conf.d/*.conf
# secure execution mode if AT_SECURE entry in auxiliary vector has nonzero value (process real & effective user IDs differ, or real & effectived group IDs differ; process with non-root user ID executed a binary that conferred capabilities to the process; nonzero value set by a Linux Security Module)
# hardware capability directories can be cascaded
# ----- hardware capabilities, recognized names -----
# Alpha: ev4, ev5, ev56, ev6, ev67
# MIPS: loongson2e, loongson2f, octeon, octeon2
# PowerPC: 4xxmac, altivec, arch_2_05, arch_2_06, booke, cellbe, dfp, efpdouble, efpsingle, fpu, ic_snoop, mmu, notb, pa6t, power4, power5, power5+, power6x, ppc32, ppc601, ppc64, smt, spe, ucache, vsx
# SPARC: flush, muldiv, stbar, swap, ultra3, v9, v9v, v9v2
# s390: dfp, eimm, esan3, etf3enh, g5, highgprs, hpage, ldisp, msa, stfle, z900, z990, z9-109, z10, zarch
# x86 (32-bit only): acpi, apic, clflush, cmov, cx8, dts, fxsr, ht, i386, i486, i586, i686, mca, mmx, mtrr, pat, pbe, pge, pn, pse36, sep, ss, sse, sse2, tm


@surfactant.plugin.hookimpl
def establish_relationships(
    sbom: SBOM, software: Software, metadata
) -> Optional[List[Relationship]]:
    if not has_required_fields(metadata):
        return None

    relationships: List[Relationship] = []
    dependent_uuid = software.UUID
    default_search_paths = generate_search_paths(software, metadata)
    for dep in metadata["elfDependencies"]:
        # if dependency has a slash, it is interpreted as a pathname to shared object to load
        # construct fname and full file path(s) to search for; paths must be a list if the dependency is given as a relative path
        if "/" in dep:
            # search SBOM entries for a library at a matching relative/absolute path
            dep = posix_normpath(
                dep
            )  # normpath takes care of redundancies such as `//`->`/` and `ab/../xy`->`xy`; NOTE may change meaning of path containing symlinks
            fname = dep.name
            if dep.is_absolute():
                # absolute path
                fpaths = [str(dep)]
            else:
                # relative path
                fpaths = []
                # iterate through install paths for sw to get the full path to the file as it would appear in installPaths for the software entry
                if isinstance(software.installPath, Iterable):
                    for ipath in software.installPath:
                        ipath_posix = posix_normpath(
                            ipath
                        )  # NOTE symlinks in install path may be affected by normpath
                        fpaths.append(
                            posix_normpath(str(ipath_posix.parent.joinpath(dep))).as_posix()
                        )  # paths to search are install path folders + relative path of dependency
        else:
            fname = dep
            # the paths for the dependency follow the default search path order for Linux/FreeBSD/etc
            fpaths = [
                p.joinpath(fname).as_posix() for p in default_search_paths
            ]  # append fname to the end of the paths to get the full file install paths of the dependency

        # Look for a software entry with a file name and install path that matches the dependency that would be loaded
        for item in sbom.software:
            # Check if the software entry has a name matching the dependency first as a quick check to rule out non-matches
            if isinstance(item.fileName, Iterable) and fname not in item.fileName:
                continue

            # check if the software entry is installed to one of the paths looked at for loading the dependency
            for fp in fpaths:
                if isinstance(item.installPath, Iterable) and fp in item.installPath:
                    # software matching requirements to be the loaded dependency was found
                    dependency_uuid = item.UUID
                    rel = Relationship(dependent_uuid, dependency_uuid, "Uses")
                    if rel not in relationships:
                        relationships.append(rel)
    return relationships


def generate_search_paths(sw: Software, md) -> List[pathlib.PurePosixPath]:
    # 1. Search using directories in DT_RPATH if present and no DT_RUNPATH exists (use of DT_RPATH is deprecated)
    # 2. Use LD_LIBRARY_PATH environment variable; ignore if suid/sgid binary (nothing to do, we don't have this information w/o running on a live system)
    # 3. Search using directories in DT_RUNPATH if present
    paths = generate_runpaths(sw, md)  # will return an empty list if none

    # 4. From /etc/ld.so.cache (/var/run/ld.so.hints on FreeBSD) list of compiled candidate libraries previously found in augmented library path; if binary was linked with -z nodeflib linker option, libraries in default library paths are skipped
    # /etc/ld.so.conf can be used to add additional directories to defaults (e.g. /usr/local/lib or /opt/lib), but we don't necessarily have a way to gather this info
    # Search in default path /lib, then /usr/lib; skip if binary was linked with -z nodeflib option
    nodeflib = False
    if "elfDynamicFlags1" in md:
        if "DF_1_NODEFLIB" in md["elfDynamicFlags1"]:
            nodeflib = md["elfDynamicFlags1"]["DF_1_NODEFLIB"]
    if not nodeflib:
        # add default search paths
        paths.extend(
            [pathlib.PurePosixPath(p) for p in ["/lib", "/lib64", "/usr/lib", "/usr/lib64"]]
        )

    return paths


def generate_runpaths(sw: Software, md) -> List[pathlib.PurePosixPath]:
    # rpath and runpath are lists of strings (just in case an ELF file has several, though that is probably an invalid ELF file)
    rp_to_use = []
    rpath = None
    runpath = None
    if "elfRpath" in md and md["elfRpath"]:
        rpath = md["elfRpath"]
    if "elfRunpath" in md and md["elfRunpath"]:
        runpath = md["elfRunpath"]

    # 1. Search using directories in DT_RPATH if present and no DT_RUNPATH exists (use of DT_RPATH is deprecated)
    # 3. Search using directories in DT_RUNPATH if present
    if rpath and not runpath:
        rp_to_use = rpath
    elif runpath:
        rp_to_use = runpath

    # split up the paths first, then substitute DSTs
    return [
        sp  # append path with DSTs replaced to the list
        for rp in rp_to_use  # iterate through all possible runpath entries
        for p in rp.split(":")  # iterate through all components (paths) in each runpath entry
        if p != ""  # if the path entry is not empty
        for sp in substitute_all_dst(sw, md, p)  # substitute DSTs in the path
    ]


def replace_dst(origstr, dvar, newval) -> str:
    return origstr.replace("$" + dvar, newval).replace("${" + dvar + "}", newval)


def substitute_all_dst(sw: Software, md, path) -> List[pathlib.PurePosixPath]:
    # substitute any dynamic string tokens found; may result in multiple strings if different variants are possible
    # replace $ORIGIN, ${ORIGIN}, $LIB, ${LIB}, $PLATFORM, ${PLATFORM} tokens
    # places the dynamic linker does this expansion are:
    # - environment vars: LD_LIBRARY_PATH, LD_PRELOAD, and LD_AUDIT
    # - dynamic section tags: DT_NEEDED, DT_RPATH, DT_RUNPATH, DT_AUDIT, and DT_DEPAUDIT
    # - arguments to ld.so: --audit, --library-path, and --preload
    # - the filename arguments to dlopen and dlmopen
    # more details in the `Dynamic string tokens` section of https://man7.org/linux/man-pages/man8/ld.so.8.html
    pathlist: List[pathlib.PurePosixPath] = []
    # ORIGIN: replace with absolute directory containing the program or shared object (with symlinks resolved and no ../ or ./ subfolders)
    # for SUID/SGID binaries, after expansion the normalized path must be in a trusted directory (https://github.com/bminor/glibc/blob/0d41182/elf/dl-load.c#L356-L357, https://github.com/bminor/glibc/blob/0d41182/elf/dl-load.c#L297-L316)
    if (path.find("$ORIGIN") != -1) or (path.find("${ORIGIN}") != -1):
        if isinstance(sw.installPath, Iterable):
            for ipath in sw.installPath:
                origin = pathlib.PurePosixPath(ipath).parent.as_posix()
                pathlist.append(pathlib.PurePosixPath(replace_dst(path, "ORIGIN", origin)))

    # LIB: expands to `lib` or `lib64` depending on arch (x86-64 to lib64, x86-32 to lib)
    if (path.find("$LIB") != -1) or (path.find("${LIB}") != -1):
        if not pathlist:
            # nothing in the original pathlist, use the original path passed in
            pathlist.append(pathlib.PurePosixPath(replace_dst(path, "LIB", "lib")))
            pathlist.append(pathlib.PurePosixPath(replace_dst(path, "LIB", "lib64")))
        else:
            # perform substitutions with every current entry in pathlist
            pathlist = [
                newp
                for p in pathlist
                for newp in (
                    pathlib.PurePosixPath(replace_dst(p, "LIB", "lib")),
                    pathlib.PurePosixPath(replace_dst(p, "LIB", "lib64")),
                )
            ]

    # PLATFORM: expands to string corresponding to CPU type of the host system (e.g. "x86_64")
    # some archs the string comes from AT_PLATFORM value in auxiliary vector (getauxval)
    if (path.find("$PLATFORM") != -1) or (path.find("${PLATFORM}") != -1):
        # NOTE consider using what is known about the target CPU of the ELF binary, and get all possible PLATFORM values based on that from glibc/muslc source code?
        #      this would take some significant amount of searching (inconsistent in how different platforms set the value), and could result in a large increase in
        #      the number of search paths for a feature that is rarely used (similar to hwcaps subfolder searching)
        # For now, discard paths given that no valid substitution was found
        return []

    # normalize paths after expanding tokens to avoid portions of the path involving  ../, ./, and // occurrences
    pathlist = [posix_normpath(p.as_posix()) for p in pathlist]
    return pathlist



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import pluggy

hookimpl = pluggy.HookimplMarker("surfactant")



# Things related to Mach-O cpu type and subtypes, pulled from:
# - https://github.com/llvm/llvm-project/blob/main/llvm/include/llvm/BinaryFormat/MachO.h
# - MacOSX14.4.sdk/usr/include/mach/machine.h (located in /Library/Developer/CommandLineTools/SDKs/ on macOS)

from enum import IntEnum

CPU_ARCH_MASK = 0xFF000000  # Mask for architecture bits
CPU_ARCH_ABI64 = 0x01000000  # 64 bit ABI
CPU_ARCH_ABI64_32 = 0x02000000  # ILP32 ABI on 64-bit hardware


# original #define for cpu types in MacOS SDK start with CPU_TYPE_
# original #define for cpu sybtypes in MacOS SDK start with CPU_SUBTYPE_
class CPU_Type(IntEnum):
    ANY = -1
    VAX = 1
    MC680x0 = 6
    X86 = 7
    I386 = X86  # compatibility
    X86_64 = X86 | CPU_ARCH_ABI64
    MIPS = 8
    MC98000 = 10  # Old Motorola PowerPC
    HPPA = 11
    ARM = 12
    ARM64 = ARM | CPU_ARCH_ABI64
    ARM64_32 = ARM | CPU_ARCH_ABI64_32
    MC88000 = 13
    SPARC = 14
    I860 = 15
    ALPHA = 16
    POWERPC = 18
    POWERPC64 = POWERPC | CPU_ARCH_ABI64


CPU_SUBTYPE_MASK = 0xFF000000  # Mask for feature flags
CPU_SUBTYPE_LIB64 = 0x80000000  # 64 bit libraries
CPU_SUBTYPE_PTRAUTH_ABI = 0x80000000  # pointer authentication with versioned ABI


class cpuSubtype_VAX(IntEnum):
    VAX_ALL = 0
    VAX780 = 1
    VAX785 = 2
    VAX750 = 3
    VAX730 = 4
    UVAXI = 5
    UVAXII = 6
    VAX8200 = 7
    VAX8500 = 8
    VAX8600 = 9
    VAX8650 = 10
    VAX8800 = 11
    UVAXIII = 12


class cpuSubtype_MC680x0(IntEnum):
    MC680x0_ALL = 1
    MC68030 = MC680x0_ALL
    MC68040 = 2
    MC68030_ONLY = 3


def CPU_SUBTYPE_INTEL(f, m):
    return f + (m << 4)


def CPU_SUBTYPE_INTEL_FAMILY(x):
    return x & 15


def CPU_SUBTYPE_INTEL_MODEL(x):
    return x >> 4


class cpuSubtype_I386(IntEnum):
    I386_ALL = CPU_SUBTYPE_INTEL(3, 0)
    _386 = I386_ALL
    _486 = CPU_SUBTYPE_INTEL(4, 0)
    _486SX = CPU_SUBTYPE_INTEL(4, 8)
    PENT = CPU_SUBTYPE_INTEL(5, 0)
    _586 = PENT
    PENTPRO = CPU_SUBTYPE_INTEL(6, 1)
    PENTII_M3 = CPU_SUBTYPE_INTEL(6, 3)
    PENTII_M5 = CPU_SUBTYPE_INTEL(6, 5)
    CELERON = CPU_SUBTYPE_INTEL(7, 6)
    CELERON_MOBILE = CPU_SUBTYPE_INTEL(7, 7)
    PENTIUM_3 = CPU_SUBTYPE_INTEL(8, 0)
    PENTIUM_3_M = CPU_SUBTYPE_INTEL(8, 1)
    PENTIUM_3_XEON = CPU_SUBTYPE_INTEL(8, 2)
    PENTIUM_M = CPU_SUBTYPE_INTEL(9, 0)
    PENTIUM_4 = CPU_SUBTYPE_INTEL(10, 0)
    PENTIUM_4_M = CPU_SUBTYPE_INTEL(10, 1)
    ITANIUM = CPU_SUBTYPE_INTEL(11, 0)
    ITANIUM_2 = CPU_SUBTYPE_INTEL(11, 1)
    XEON = CPU_SUBTYPE_INTEL(12, 0)
    XEON_MP = CPU_SUBTYPE_INTEL(12, 1)


class cpuSubtype_X86(IntEnum):
    X86_ALL = 3
    X86_64_ALL = 3
    X86_64_ARCH1 = 4
    X86_64_H = 8  # Haswell feature subset


class cpuSubtype_MIPS(IntEnum):
    MIPS_ALL = 0
    MIPS_R2300 = 1
    MIPS_R2600 = 2
    MIPS_R2800 = 3
    MIPS_R2000a = 4  # pmax
    MIPS_R2000 = 5
    MIPS_R3000a = 6  # 3max
    MIPS_R3000 = 7


class cpuSubtype_MC98000(IntEnum):
    MC980000_ALL = 0
    MC98601 = 1


# Hewlett-Packard HP-PA processors
class cpuSubtype_HPPA(IntEnum):
    HPPA_ALL = 0
    HPPA_7100 = HPPA_ALL  # compatibility
    HPPA_7100LC = 1


class cpuSubtype_MC88000(IntEnum):
    MC88000_ALL = 0
    MC88100 = 1
    MC88110 = 2


class cpuSubtype_SPARC(IntEnum):
    SPARC_ALL = 0


class cpuSubtype_I860(IntEnum):
    I860_ALL = 0
    I860_860 = 1


class cpuSubtype_PowerPC(IntEnum):
    POWERPC_ALL = 0
    POWERPC_601 = 1
    POWERPC_602 = 2
    POWERPC_603 = 3
    POWERPC_603e = 4
    POWERPC_603ev = 5
    POWERPC_604 = 6
    POWERPC_604e = 7
    POWERPC_620 = 8
    POWERPC_750 = 9
    POWERPC_7400 = 10
    POWERPC_7450 = 11
    POWERPC_970 = 100


class cpuSubtype_ARM(IntEnum):
    ARM_ALL = 0
    ARM_V4T = 5
    ARM_V6 = 6
    ARM_V5TEJ = 7  # also ARM_V5TEJ
    ARM_XSCALE = 8
    ARM_V7 = 9  # ARMv7-A and ARMv7-R
    ARM_V7F = 10  # Cortex A8
    ARM_V7S = 11  # Swift
    ARM_V7K = 12
    ARM_V8 = 13
    ARM_V6M = 14
    ARM_V7M = 15
    ARM_V7EM = 16
    ARM_V8M = 17


class cpuSubtype_ARM64(IntEnum):
    ARM64_ALL = 0
    ARM64_V8 = 1
    ARM64E = 2


class cpuSubtype_ARM64_32(IntEnum):
    ARM64_32_ALL = 0
    ARM64_32_V8 = 1


def get_cpu_subtype_name(cpuType: CPU_Type, cpuSubtype) -> str:
    cpu_subtype_mapping = {
        CPU_Type.VAX: cpuSubtype_VAX,
        CPU_Type.MC680x0: cpuSubtype_MC680x0,
        CPU_Type.X86: cpuSubtype_I386,
        CPU_Type.X86_64: cpuSubtype_X86,
        CPU_Type.MIPS: cpuSubtype_MIPS,
        CPU_Type.MC98000: cpuSubtype_MC98000,
        CPU_Type.HPPA: cpuSubtype_HPPA,
        CPU_Type.ARM: cpuSubtype_ARM,
        CPU_Type.ARM64: cpuSubtype_ARM64,
        CPU_Type.ARM64_32: cpuSubtype_ARM64_32,
        CPU_Type.MC88000: cpuSubtype_MC88000,
        CPU_Type.SPARC: cpuSubtype_SPARC,
        CPU_Type.I860: cpuSubtype_I860,
        CPU_Type.POWERPC: cpuSubtype_PowerPC,
        CPU_Type.POWERPC64: cpuSubtype_PowerPC,  # Does PowerPC64 use the same subtypes as PowerPC?
    }

    subtype_enum = cpu_subtype_mapping.get(cpuType)

    if subtype_enum is None:
        return "UNKNOWN"

    try:
        subtype_name = subtype_enum(cpuSubtype).name
        return subtype_name
    except ValueError:
        return "UNKNOWN"


def get_cpu_type_name(cpuType: CPU_Type) -> str:
    if isinstance(cpuType, int):
        try:
            cpuType = CPU_Type(cpuType)
        except ValueError:
            return "UNKNOWN"

    if isinstance(cpuType, CPU_Type):
        return cpuType.name
    return "UNKNOWN"



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import pathlib
from collections.abc import Iterable
from typing import Any, List, Union

from surfactant.sbomtypes import SBOM, Software


# return all matching dotnet assemblies or DLLs that could be loaded on Windows
# TODO: an intermediate file format should keep files in different places but matching hashes separate until
# relationships are established; this would make so we can use .NET metadata about versions, strong names, etc
# and not accidentally mix and match cultures/app config info that could differ for different copies of the same
# file (due to app config files pointing to different assemblies despite DLL having same hash)
# culture information to find the right assembly from app config file is likely to vary (though almost always neutral/none)
def find_installed_software(
    sbom: SBOM, probedirs: List[Any], filename: Union[str, List[str]]
) -> List[Software]:
    possible_matches = []
    # iterate through all sbom entries
    for e in sbom.software:
        # Skip if no install path (e.g. installer/temporary file)
        if e.installPath is None:
            continue
        for pdir in probedirs:
            if isinstance(filename, str):
                filename = [filename]
            for fname in filename:
                # installPath contains full path+filename, so check for all combinations of probedirs+filename
                pfile = pathlib.PureWindowsPath(pdir, fname)
                if isinstance(e.installPath, Iterable):
                    for ifile in e.installPath:
                        # PureWindowsPath is case-insensitive for file/directory names
                        if pfile == pathlib.PureWindowsPath(ifile):
                            # matching probe directory and filename, add software to list
                            possible_matches.append(e)
    return possible_matches



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

# https://en.wikipedia.org/wiki/Comparison_of_executable_file_formats

import importlib.metadata
import sys

import click
from loguru import logger

from surfactant.cmd.cli import (
    handle_cli_add,
    handle_cli_edit,
    handle_cli_find,
    handle_cli_load,
    handle_cli_save,
)
from surfactant.cmd.config import config
from surfactant.cmd.config_tui import config_tui
from surfactant.cmd.createconfig import create_config
from surfactant.cmd.generate import sbom as generate
from surfactant.cmd.merge import merge_command
from surfactant.cmd.plugin import (
    plugin_disable_cmd,
    plugin_enable_cmd,
    plugin_install_cmd,
    plugin_list_cmd,
    plugin_uninstall_cmd,
    plugin_update_db_cmd,
)
from surfactant.cmd.stat import stat
from surfactant.cmd.ui import ui


@click.group()
@click.version_option(
    importlib.metadata.version("surfactant"),
    "--version",
    "-v",
    message="%(version)s",
)
@click.option(
    "--log-level",
    type=click.Choice(
        ["TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL"],
        case_sensitive=False,
    ),
    default="INFO",
)
def main(log_level="INFO"):
    # Can't change the logging level; need to remove and add a new logger with the desired log level
    logger.remove()
    logger.add(sys.stderr, level=log_level)


@click.command("version")
def version():
    """Print version information."""
    click.echo(importlib.metadata.version("surfactant"))
    sys.exit(0)


@main.group("cli")
def cli():
    """Commandline interface used to modify SBOM entries."""


@main.group("plugin")
def plugin():
    """Manage plugins."""


# Main Commands
main.add_command(generate)
main.add_command(version)
main.add_command(config)
main.add_command(stat)
main.add_command(merge_command)
main.add_command(create_config)
main.add_command(ui)
main.add_command(plugin)
main.add_command(config_tui)

# CLI Subcommands
cli.add_command(handle_cli_find)
cli.add_command(handle_cli_edit)
cli.add_command(handle_cli_add)
cli.add_command(handle_cli_load)
cli.add_command(handle_cli_save)

# Plugin Subcommands
plugin.add_command(plugin_list_cmd)
plugin.add_command(plugin_enable_cmd)
plugin.add_command(plugin_disable_cmd)
plugin.add_command(plugin_install_cmd)
plugin.add_command(plugin_uninstall_cmd)
plugin.add_command(plugin_update_db_cmd)

if __name__ == "__main__":
    main()



# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import json
import re
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional

from loguru import logger

import surfactant.plugin
from surfactant.configmanager import ConfigManager
from surfactant.database_manager.database_utils import (
    calculate_hash,
    download_database,
    load_hash_and_timestamp,
    save_hash_and_timestamp,
)
from surfactant.sbomtypes import SBOM, Software

# Global configuration
DATABASE_URL = "https://raw.githubusercontent.com/RetireJS/retire.js/master/repository/jsrepository-master.json"


class JSDatabaseManager:
    def __init__(self):
        self._js_lib_database: Optional[Dict[str, Any]] = None  # Use the private attribute
        self.database_version_file_path = (
            ConfigManager().get_data_dir_path()
            / "infoextractors"
            / "js_library_patterns"
            / "js_library_patterns.toml"
        )
        self.pattern_key = "js_library_patterns"
        self.pattern_file = "js_library_patterns.json"
        self.source = "jsfile.retirejs"
        self.new_hash: Optional[str] = None
        self.download_timestamp: Optional[datetime] = None

    @property
    def js_lib_database(self) -> Optional[Dict[str, Any]]:
        if self._js_lib_database is None:
            self.load_db()
        return self._js_lib_database

    @property
    def pattern_info(self) -> Dict[str, Any]:
        return {
            "pattern_key": self.pattern_key,
            "pattern_file": self.pattern_file,
            "source": self.source,
            "hash_value": self.new_hash,
            "timestamp": self.download_timestamp,
        }

    def load_db(self) -> None:
        js_lib_file = (
            ConfigManager().get_data_dir_path()
            / "infoextractors"
            / "js_library_patterns"
            / self.pattern_file
        )

        try:
            with open(js_lib_file, "r") as regex:
                self._js_lib_database = json.load(regex)
        except FileNotFoundError:
            logger.warning(
                "Javascript library pattern database could not be loaded. Run `surfactant plugin update-db js_file` to fetch the pattern database."
            )
            self._js_lib_database = None

    def get_database(self) -> Optional[Dict[str, Any]]:
        return self._js_lib_database


js_db_manager = JSDatabaseManager()


def supports_file(filetype) -> bool:
    return filetype == "JAVASCRIPT"


@surfactant.plugin.hookimpl
def extract_file_info(sbom: SBOM, software: Software, filename: str, filetype: str) -> object:
    if not supports_file(filetype):
        return None
    return extract_js_info(filename)


def extract_js_info(filename: str) -> object:
    js_info: Dict[str, Any] = {"jsLibraries": []}
    js_lib_database = js_db_manager.get_database()

    if js_lib_database is None:
        return None

    # Try to match file name
    libs = match_by_attribute("filename", filename, js_lib_database)
    if len(libs) > 0:
        js_info["jsLibraries"] = libs
        return js_info

    # Try to match file contents
    try:
        with open(filename, "r") as js_file:
            filecontent = js_file.read()
        libs = match_by_attribute("filecontent", filecontent, js_lib_database)
        js_info["jsLibraries"] = libs
    except FileNotFoundError:
        logger.warning(f"File not found: {filename}")
    except UnicodeDecodeError:
        logger.warning(f"File does not appear to be UTF-8: {filename}")
    return js_info


def match_by_attribute(attribute: str, content: str, database: Dict) -> List[Dict]:
    libs = []
    for name, library in database.items():
        if attribute in library:
            for pattern in library[attribute]:
                matches = re.search(pattern, content)
                if matches:
                    if len(matches.groups()) > 0:
                        libs.append({"library": name, "version": matches.group(1)})
                        # skip remaining patterns, move on to the next library
                        break
    return libs


def strip_irrelevant_data(retirejs_db: dict) -> dict:
    clean_db = {}
    reg_temp = "\u00a7\u00a7version\u00a7\u00a7"
    version_regex = r"\d+(?:\.\d+)*"
    for library, lib_entry in retirejs_db.items():
        if "extractors" in lib_entry:
            clean_db[library] = {}
            patterns = lib_entry["extractors"]
            possible_entries = [
                "filename",
                "filecontent",
                "hashes",
            ]
            for entry in possible_entries:
                if entry in patterns:
                    entry_list = []
                    for reg in patterns[entry]:
                        entry_list.append(reg.replace(reg_temp, version_regex))
                    clean_db[library][entry] = entry_list
    return clean_db


@surfactant.plugin.hookimpl
def update_db() -> str:
    raw_data = download_database(DATABASE_URL)
    if raw_data is not None:
        js_db_manager.new_hash = calculate_hash(raw_data)
        current_data = load_hash_and_timestamp(
            js_db_manager.database_version_file_path,
            js_db_manager.pattern_key,
            js_db_manager.pattern_file,
        )
        if current_data and js_db_manager.new_hash == current_data.get("hash"):
            return "No update occurred. Database is up-to-date."

        retirejs = json.loads(raw_data)
        cleaned = strip_irrelevant_data(retirejs)
        js_db_manager.download_timestamp = datetime.now(timezone.utc)

        path = ConfigManager().get_data_dir_path() / "infoextractors" / "js_library_patterns"
        path.mkdir(parents=True, exist_ok=True)
        json_file_path = path / js_db_manager.pattern_file
        with open(json_file_path, "w") as f:
            json.dump(cleaned, f, indent=4)

        save_hash_and_timestamp(
            js_db_manager.database_version_file_path, js_db_manager.pattern_info
        )
        return "Update complete."
    return "No update occurred."


@surfactant.plugin.hookimpl
def short_name() -> str:
    return "js_file"


@surfactant.plugin.hookimpl
def init_hook(command_name: Optional[str] = None):
    """
    Initialization hook to load the JavaScript library database.

    Args:
        command_name (Optional[str], optional): The name of the command invoking the initialization.
            If set to "update-db", the database will not be loaded.

    Returns:
        None
    """
    if command_name != "update-db":  # Do not load the database if only updating the database.
        logger.info("Initializing js_file...")
        js_db_manager.load_db()
        logger.info("Initializing js_file complete.")



import os
import sys

import requests

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Surfactant"
# pylint: disable-next=redefined-builtin
copyright = "2023, Lawrence Livermore National Security"
author = "Ryan Mast, Kendall Harter, Micaela Gallegos, Shayna Kapadia, Apoorv Pochiraju, Alexander Armstrong, Levi Lloyd"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx.ext.githubpages",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "images.toml"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "alabaster"
html_theme_options = {
    "description": "Surfactant",
    "github_user": "LLNL",
    "github_repo": "Surfactant",
    "github_button": "true",
    "github_banner": "true",
    "badge_branch": "main",
    "fixed_sidebar": "false",
}

# -- Extension configuration -------------------------------------------------

# Napoleon settings for NumPy and Google style docstrings
napoleon_google_docstring = True
napoleon_numpy_docstring = True
html_logo = "./logos/surfactant-logo-light.png"
html_favicon = html_logo
html_sidebars = {"**": ["globaltoc.html", "relations.html", "searchbox.html"]}
html_static_path = ["_static"]


# -- Fetch image references --------------------------------------------------
# Download all of the image files referenced in images.toml
def download_images_from_toml(toml_file, image_dir):
    with open(toml_file, "rb") as f:
        data = tomllib.load(f)

    if not os.path.exists(image_dir):
        os.makedirs(image_dir)

    for file_name, url in data.get("images", {}).items():
        if file_name and url:
            response = requests.get(url)
            if response.status_code == 200:
                with open(os.path.join(image_dir, file_name), "wb") as img_file:
                    img_file.write(response.content)
            else:
                print(f"Failed to download {url}")


# Path to the TOML file
toml_file_path = os.path.join(os.path.dirname(__file__), "images.toml")
# Directory to save the images
image_directory = os.path.join(os.path.dirname(__file__), "img")

# Download images
download_images_from_toml(toml_file_path, image_directory)



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import argparse
import json
import sys
import uuid as uuid_module
from collections import deque
from typing import Optional


def is_valid_uuid4(u) -> bool:
    """Check if a uuid is valid

    Args:
        u (string): uuid in string form.

    Returns:
        bool: true if u is a valid uuid and false if it is not.
    """
    try:
        u_test = uuid_module.UUID(u, version=4)
    except ValueError:
        return False
    return str(u_test) == u


def find_relationship_entry(
    sbom,
    xUUID: Optional[str] = None,
    yUUID: Optional[str] = None,
    relationship: Optional[str] = None,
) -> Optional[dict]:
    """Search for a specific relationship entry and check if a match exists.

    Args:
        sbom (dict): Dictionary containing sbom entries.
        xUUID (Optional[str]): Component x UUID. Defaults to None.
        yUUID (Optional[str]): Component y UUID. Defaults to None.
        relationship (Optional[str]): Describes the relationships between two components. Options are 'Uses', 'Contains'. Defaults to None.

    Returns:
        Optional[dict]: Dictionary entry that contains information where the xUUID, yUUID and relationship all match. If no match, returns None.
    """
    for rel in sbom["relationships"]:
        all_match = True
        if xUUID:
            if rel["xUUID"] != xUUID:
                all_match = False
        if yUUID:
            if rel["yUUID"] != yUUID:
                all_match = False
        if relationship:
            if rel["relationship"] != relationship:
                all_match = False
        if all_match:
            return rel
    return None


def find_star_relationship_entry(
    sbom,
    xUUID: Optional[str] = None,
    yUUID: Optional[str] = None,
    relationship: Optional[str] = None,
) -> Optional[dict]:
    """Search for a star relationship entry and check if a match exists.

    Args:
        sbom (dict): Dictionary containing sbom entries.
        xUUID (Optional[str]): Component x UUID. Defaults to None.
        yUUID (Optional[str]): Component y UUID. Defaults to None.
        relationship (Optional[str]): Describes the star relationship between two components. Defaults to None.

    Returns:
        Optional[dict]: Dictionary entry that contains information wehre the xUUID, yUUID, and relationship all match. If no match, returns None.
    """
    for rel in sbom["starRelationships"]:
        all_match = True
        if xUUID:
            if rel["xUUID"] != xUUID:
                all_match = False
        if yUUID:
            if rel["yUUID"] != yUUID:
                all_match = False
        if relationship:
            if rel["relationship"] != relationship:
                all_match = False
        if all_match:
            return rel
    return None


def find_systems_entry(
    sbom, uuid: Optional[str] = None, name: Optional[str] = None
) -> Optional[dict]:
    """Search for a systems entry and check if a match exists.

    Args:
        sbom (dict): Dictionary containing sbom entries.
        uuid (Optional[str]): Contains component UUID. Defaults to None.
        name (Optional[str]): Name of the larger file the component came from. Defaults to None.

    Returns:
        Optional[dict]: Dictionary entry that contains the matching system entry. If no match, returns None.
    """
    for system in sbom["systems"]:
        all_match = True
        if uuid:
            if system["UUID"] != uuid:
                all_match = False
        if name:
            if system["name"] != name:
                all_match = False
        if all_match:
            return system
    return None


def find_software_entry(
    sbom,
    uuid: Optional[str] = None,
    sha256: Optional[str] = None,
    md5: Optional[str] = None,
    sha1: Optional[str] = None,
) -> Optional[dict]:
    """Search for a specific software entry and check if a match exists.

    Args:
        sbom (dict): Dictionary containing sbom entries.
        uuid (Optional[str]): Contains component UUID. Defaults to None.
        sha256 (Optional[str]): SHA256 hash of component. Defaults to None.
        md5 (Optional[str]): MD5 hash of component. Defaults to None.
        sha1 (Optional[str]): SHA1 hash of component. Defaults to None.

    Returns:
        Optional[dict]: Dictionary entry that contains the matching software entry. If no match, returns None.
    """
    for sw in sbom["software"]:
        all_match = True
        if uuid:
            if sw["UUID"] != uuid:
                all_match = False
        if sha256:
            if sw["sha256"] != sha256:
                all_match = False
        if md5:
            if sw["md5"] != md5:
                all_match = False
        if sha1:
            if sw["sha1"] != sha1:
                all_match = False
        if all_match:
            return sw
    return None


def merge_number_same(e1: int, e2: int, k: str):
    """Merges two entries if the number is the same.

    Args:
        e1 (int): Number.
        e2 (int): Number.
        k (str): Name of the field to compare the two numbers.
    """
    # use e2 number if field isn't in e1
    if k not in e1:
        if k in e2:
            e1[k] = e2[k]
    else:
        if k in e2:
            if e1[k] != e2[k]:
                print(f"Field {k} that should match does not! e1={e1} e2={e2}")


def merge_number_lt(e1: int, e2: int, k: str):
    """Merge two entries and assign the lesser of the two numbers.

    Args:
        e1 (int): Number.
        e2 (int): Number.
        k (str): Name of the field to compare the two numbers.
    """
    # use e2 number if the field isn't in e1
    if k not in e1:
        if k in e2:
            e1[k] = e2[k]
    else:
        if k in e2:
            if e1[k] > e2[k]:
                e1[k] = e2[k]


def merge_number_gt(e1: int, e2: int, k: str):
    """Merge two entries and assign the greater of the two numbers.

    Args:
        e1 (int): Number.
        e2 (int): Number.
        k (str): Name of the field to compare the two numbers.
    """
    # use e2 number if the field isn't in e1
    if k not in e1:
        if k in e2:
            e1[k] = e2[k]
    else:
        if k in e2:
            if e1[k] < e2[k]:
                e1[k] = e2[k]


def merge_string(e1: str, e2: str, k: str):
    """Merge two entries. If empty, keep empty.

    Args:
        e1 (str): String with information.
        e2 (str): String with information.
        k (str): Name of the field to compare the two strings.
    """
    if k not in e1 or not e1[k]:
        # worst case, e2 has an empty string/null just like e1
        if k in e2:
            e1[k] = e2[k]


def merge_array(e1, e2, k):
    if k not in e1 or not e1[k]:
        # e2 is at worst empty list like e1, or contains some entries
        if k in e2:
            e1[k] = e2[k]
    else:
        # e1 is guaranteed to have a list with at least one entry
        # merge in items from e2 that aren't already in e1
        if k in e2 and e2[k]:
            for item in e2[k]:
                if item not in e1[k]:
                    e1[k].append(item)


# merges data from two systems entries, modifying the first to contain new info from the second
def merge_systems_entries(e1, e2):
    # name
    merge_string(e1, e2, "name")
    # captureStart
    merge_number_lt(e1, e2, "captureStart")
    # captureEnd
    merge_number_gt(e1, e2, "captureEnd")
    # officialName
    merge_string(e1, e2, "officialName")
    # vendor
    merge_array(e1, e2, "vendor")
    # description
    merge_string(e1, e2, "description")
    # provenance
    merge_array(e1, e2, "provenance")
    return e1["UUID"], e2["UUID"]


# merges data from two software entries, modifying the first entry to contain new info from the second (under the assumption that hashes are the same, so certain values must match)
def merge_software_entries(e1, e2):
    # sha256
    merge_string(e1, e2, "sha256")
    # sha1
    merge_string(e1, e2, "sha1")
    # md5
    merge_string(e1, e2, "md5")
    # name
    merge_string(e1, e2, "name")
    # fileName
    merge_array(e1, e2, "fileName")
    # installPath
    merge_array(e1, e2, "installPath")
    # containerPath
    merge_array(e1, e2, "containerPath")
    # size
    merge_number_same(e1, e2, "size")
    # captureTime
    merge_number_lt(e1, e2, "captureTime")  # favor the older time
    # version
    merge_string(e1, e2, "version")
    # vendor
    merge_array(e1, e2, "vendor")
    # description
    merge_string(e1, e2, "description")
    # relationshipAssertion
    merge_string(e1, e2, "relationshipAssertion")
    if "relationshipAssertion" in e1 and e1["relationshipAssertion"] == "Unknown":
        # e2 has a better relationshipAssertion than "Unknown"
        if (
            "relationshipAssertion" in e2
            and e2["relationshipAssertion"]
            and e2["relationshipAssertion"] != "Unknown"
        ):
            e1["relationshipAssertion"] = e2["relationshipAssertion"]
    # comments
    merge_string(e1, e2, "comments")
    # metadata
    merge_array(e1, e2, "metadata")
    # supplementaryFiles
    merge_array(e1, e2, "supplementaryFiles")
    # provenance
    merge_array(e1, e2, "provenance")
    # recordedInstitution
    merge_string(e1, e2, "recordedInstitution")
    # components
    merge_array(e1, e2, "components")
    return e1["UUID"], e2["UUID"]


def merge_sbom(sbom_m, sbom):
    # some older SBOMs might have duplicate hashes within them... create a new merged sbom to avoid issues...
    merged_sbom = {
        "systems": [],
        "software": [],
        "relationships": [],
        "analysisData": [],
        "observations": [],
        "starRelationships": [],
    }
    # merged/old to new UUID map
    uuid_updates = {}
    # merge systems entries
    if "systems" in sbom_m:
        for system in sbom_m["systems"]:
            # check for duplicate UUID/name, merge with existing entry
            if existing_system := find_systems_entry(
                merged_sbom, uuid=system["UUID"], name=system["name"]
            ):
                # merge system entries
                u1, u2 = merge_systems_entries(existing_system, system)
                print(f"MERGE_DUPLICATE_SYS: uuid1={u1}, uuid2={u2}")
                uuid_updates[u2] = u1
            else:
                merged_sbom["systems"].append(system)
    if "systems" in sbom:
        for system in sbom["systems"]:
            # check for duplicate UUID/name, merge with existing entry
            if existing_system := find_systems_entry(
                merged_sbom, uuid=system["UUID"], name=system["name"]
            ):
                # merge system entries
                u1, u2 = merge_systems_entries(existing_system, system)
                print(f"MERGE_DUPLICATE_SYS: uuid1={u1}, uuid2={u2}")
                uuid_updates[u2] = u1
            else:
                merged_sbom["systems"].append(system)
    # merge software entries
    if "software" in sbom_m:
        for sw in sbom_m["software"]:
            # check for a duplicate hash, and merge with the existing entry
            if existing_sw := find_software_entry(
                merged_sbom, sha256=sw["sha256"], md5=sw["md5"], sha1=sw["sha1"]
            ):
                u1, u2 = merge_software_entries(existing_sw, sw)
                print(f"MERGE_DUPLICATE: uuid1={u1}, uuid2={u2}")
                uuid_updates[u2] = u1
            else:
                merged_sbom["software"].append(sw)
    if "software" in sbom:
        for sw in sbom["software"]:
            if existing_sw := find_software_entry(
                merged_sbom, sha256=sw["sha256"], md5=sw["md5"], sha1=sw["sha1"]
            ):
                u1, u2 = merge_software_entries(existing_sw, sw)
                print(f"MERGE DUPLICATE: uuid1={u1}, uuid2={u2}")
                uuid_updates[u2] = u1
            else:
                merged_sbom["software"].append(sw)
    # merge relationships
    if "relationships" in sbom_m:
        for rel in sbom_m["relationships"]:
            # rewrite UUIDs before doing the search
            if rel["xUUID"] in uuid_updates:
                rel["xUUID"] = uuid_updates[rel["xUUID"]]
            if rel["yUUID"] in uuid_updates:
                rel["yUUID"] = uuid_updates[rel["xUUID"]]
            if existing_rel := find_relationship_entry(
                merged_sbom,
                xUUID=rel["xUUID"],
                yUUID=rel["yUUID"],
                relationship=rel["relationship"],
            ):
                print("DUPLICATE RELATIONSHIP: {existing_rel}")
            else:
                merged_sbom["relationships"].append(rel)
    if "relationships" in sbom:
        for rel in sbom["relationships"]:
            # rewrite UUIDs before doing the search
            if rel["xUUID"] in uuid_updates:
                rel["xUUID"] = uuid_updates[rel["xUUID"]]
            if rel["yUUID"] in uuid_updates:
                rel["yUUID"] = uuid_updates[rel["yUUID"]]
            if existing_rel := find_relationship_entry(
                merged_sbom,
                xUUID=rel["xUUID"],
                yUUID=rel["yUUID"],
                relationship=rel["relationship"],
            ):
                print(f"DUPLICATE RELATIONSHIP: {existing_rel}")
            else:
                merged_sbom["relationships"].append(rel)
    # rewrite container path UUIDs using rewrite map/list
    for sw in merged_sbom["software"]:
        if "containerPath" in sw and sw["containerPath"]:
            for idx, path in enumerate(sw["containerPath"]):
                u = path[:36]
                # if container path starts with an invalid uuid4, sbom might not be valid
                if is_valid_uuid4(u):
                    if u in uuid_updates:
                        updated_path = path.replace(u, uuid_updates[u], 1)
                        sw["containerPath"][idx] = updated_path
            # remove duplicates
            sw["containerPath"] = [*set(sw["containerPath"])]
    print(f"UUID UPDATES: {uuid_updates}")
    # merge analysisData
    if "analysisData" in sbom_m:
        for analysisData in sbom_m["analysisData"]:
            # checks for duplicates might be nice, but unlikely to encounter often as these are added manually
            merged_sbom["analysisData"].append(analysisData)
    if "analysisData" in sbom:
        for analysisData in sbom["analysisData"]:
            merged_sbom["analysisData"].append(analysisData)
    # merge observations
    if "observations" in sbom_m:
        for observation in sbom_m["observations"]:
            # checks for duplicates might be nice, but unlikely to encounter often as these are added manually
            merged_sbom["observations"].append(observation)
    if "observations" in sbom:
        for observation in sbom["observations"]:
            merged_sbom["observations"].append(observation)
    # merge starRelationships
    if "starRelationships" in sbom_m:
        for rel in sbom_m["starRelationships"]:
            # rewrite UUIDs before doing the search
            if rel["xUUID"] in uuid_updates:
                rel["xUUID"] = uuid_updates[rel["xUUID"]]
            if rel["yUUID"] in uuid_updates:
                rel["yUUID"] = uuid_updates[rel["xUUID"]]
            if existing_rel := find_star_relationship_entry(
                merged_sbom,
                xUUID=rel["xUUID"],
                yUUID=rel["yUUID"],
                relationship=rel["relationship"],
            ):
                print("DUPLICATE STAR RELATIONSHIP: {existing_rel}")
            else:
                merged_sbom["starRelationships"].append(rel)
    if "starRelationships" in sbom:
        for rel in sbom["starRelationships"]:
            # rewrite UUIDs before doing the search
            if rel["xUUID"] in uuid_updates:
                rel["xUUID"] = uuid_updates[rel["xUUID"]]
            if rel["yUUID"] in uuid_updates:
                rel["yUUID"] = uuid_updates[rel["yUUID"]]
            if existing_rel := find_star_relationship_entry(
                merged_sbom,
                xUUID=rel["xUUID"],
                yUUID=rel["yUUID"],
                relationship=rel["relationship"],
            ):
                print(f"DUPLICATE RELATIONSHIP: {existing_rel}")
            else:
                merged_sbom["starRelationships"].append(rel)
    return merged_sbom


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config_file",
        metavar="CONFIG_FILE",
        nargs="?",
        type=argparse.FileType("r"),
        help="Config file (JSON); make sure keys with paths do not have a trailing /",
    )
    parser.add_argument(
        "--sbom_outfile",
        metavar="SBOM_OUTPUT",
        nargs="?",
        type=argparse.FileType("w"),
        default=sys.stdout,
        help="Output SBOM file",
    )
    parser.add_argument("input_sbom", type=argparse.FileType("r"), nargs="+")
    args = parser.parse_args()

    msbom = {
        "systems": [],
        "software": [],
        "relationships": [],
        "analysisData": [],
        "observations": [],
        "starRelationships": [],
    }
    for f in args.input_sbom:
        in_sbom = json.load(f)
        msbom = merge_sbom(msbom, in_sbom)

    print(len(msbom["software"]))
    print(len(msbom["relationships"]))

    # construct a graph for adding a system relationship to all root software entries
    rel_graph = {}
    # add all UUIDs as nodes in the graph
    for system in msbom["systems"]:
        rel_graph[system["UUID"]] = []
    for sw in msbom["software"]:
        rel_graph[sw["UUID"]] = []
    # iterate through all relationships, adding edges to the adjacency list
    for rel in msbom["relationships"]:
        # check case where xUUID doesn't exist (and error if yUUID doesn't exist) in the graph
        if rel["xUUID"] not in rel_graph or rel["yUUID"] not in rel_graph:
            print("====ERROR xUUID or yUUID doesn't exist====")
            print(rel)
            continue
        # consider also including relationship type for the edge
        # treat as directed graph, with inverted edges (pointing to parents) so dfs will eventually lead to the root parent node for a (sub)graph
        rel_graph[rel["yUUID"]].append(rel["xUUID"])

    visited = set()
    roots = set()
    rootFound = set()
    recursionStack = deque()

    # maintain a recursion stack to check for cycles; if we are visiting a node that is in the stack, there is a cycle; arbitrarily pick one to add as a root
    def dfs(rel):
        recursionStack.append(rel)
        # if the node is already visited, no revisiting required
        if rel in visited:
            recursionStack.pop()
            return rel in rootFound
        # mark the node as visited
        visited.add(rel)
        # the node has no parents, it is a root
        if not rel_graph[rel]:
            roots.add(rel)
            rootFound.add(rel)
            recursionStack.pop()
            return True
        cycle = False
        # node is not a root, move on to parents
        for parent in rel_graph[rel]:
            # detect cycles
            if parent in recursionStack:
                print(f"CYCLE DETECTED: {parent} {rel}")
                cycle = True
            if dfs(parent):
                rootFound.add(rel)
        # if there was a cycle, and none of the parents led to a definite root node
        if cycle and rel not in rootFound:
            print(f"CYCLE AND NO ROOT FOUND, SETTING {rel} AS THE ROOT")
            roots.add(rel)
            rootFound.add(rel)
        recursionStack.pop()
        return rel in rootFound

    for rel in rel_graph:
        dfs(rel)
    print(f"ROOTS: {roots}")

    # construct the most accurate system object we can, using a mix of user provided info and times loaded from the SBOM
    system = {}
    if args.config_file:
        config = json.load(args.config_file)
        system = config["system"]
    # make sure the required fields are present and at least mostly valid
    if "UUID" not in system or not is_valid_uuid4(system["UUID"]):
        system["UUID"] = str(uuid_module.uuid4())
    if "name" not in system:
        system["name"] = ""
    captureStart = -1
    captureEnd = -1
    for sw in msbom["software"]:
        if captureStart == -1 or sw["captureTime"] < captureStart:
            captureStart = sw["captureTime"]
        if captureEnd == -1 or sw["captureTime"] > captureEnd:
            captureEnd = sw["captureTime"]
    if "captureStart" not in system or not system["captureStart"]:
        system["captureStart"] = captureStart
    if "captureEnd" not in system or not system["captureEnd"]:
        system["captureEnd"] = captureEnd
    msbom["systems"].append(system)
    print(f"SYSTEM: {system}")

    # add a system relationship to each root software/systems entry identified
    for r in roots:
        msbom["relationships"].append(
            {"xUUID": system["UUID"], "yUUID": r, "relationship": "Includes"}
        )

    json.dump(msbom, args.sbom_outfile, indent=4)


if __name__ == "__main__":
    main()



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import json
from pathlib import Path

import angr
from cle import CLECompatibilityError
from loguru import logger

import surfactant.plugin
from surfactant.sbomtypes import SBOM, Software


@surfactant.plugin.hookimpl(specname="extract_file_info")
# extract_strings(sbom: SBOM, software: Software, filename: str, filetype: str):
# def angrimport_finder(filename: str, filetype: str, filehash: str):
def angrimport_finder(sbom: SBOM, software: Software, filename: str, filetype: str):
    """
    :param sbom(SBOM): The SBOM that the software entry/file is being added to. Can be used to add observations or analysis data.
    :param software(Software): The software entry associated with the file to extract information from.
    :param filename (str): The full path to the file to extract information from.
    :param filetype (str): File type information based on magic bytes.
    """

    # Only parsing executable files
    if filetype not in ["ELF", "PE"]:
        pass
    filehash = str(software.sha256)
    filename = Path(filename)
    flist = []

    # Performing check to see if file has been analyzed already
    existing_json = None
    output_name = None
    for f in Path.cwd().glob("*_additional_metadata.json"):
        flist.append((f.stem).split("_")[0])
        if filehash == (f.stem).split("_")[0]:
            existing_json = f
            output_name = f

    if existing_json:
        with open(existing_json, "r") as json_file:
            existing_data = json.load(json_file)
        if "imported function names" in existing_data:
            logger.info(f"Already extracted {filename.name}")
        else:
            try:
                logger.info(
                    f"Found existing JSON file for {filename.name} but without 'imported functions' key. Proceeding with extraction."
                )
                # Add your extraction code here.
                if filename.name not in existing_data["filename"]:
                    existing_data["filename"].append(filename.name)
                existing_data["imported function names"] = []
                # Create an angr project
                project = angr.Project(filename, auto_load_libs=False)

                # Get the imported functions using symbol information
                for symbol in project.loader.main_object.symbols:
                    if symbol.is_function:
                        existing_data["imported function names"].append(symbol.name)

                # Write the string_dict to the output JSON file
                with open(output_name, "w") as json_file:
                    json.dump(existing_data, json_file, indent=4)
            except CLECompatibilityError as e:
                logger.info(f"Angr Error {filename} {e}")
    else:
        try:
            # Validate the file path
            if not filename.exists():
                raise FileNotFoundError(f"No such file: '{filename}'")

            # Extract filename without extension
            output_path = Path.cwd() / f"{filehash}_additional_metadata.json"
            metadata = {}
            metadata["sha256hash"] = filehash
            metadata["filename"] = [filename.name]
            metadata["imported function names"] = []
            # Create an angr project
            project = angr.Project(filename.as_posix(), auto_load_libs=False)
            # Get the imported functions using symbol information
            for symbol in project.loader.main_object.symbols:
                if symbol.is_function:
                    metadata["imported function names"].append(symbol.name)

            # Write the string_dict to the output JSON file
            with open(output_path, "w") as json_file:
                json.dump(metadata, json_file, indent=4)

            logger.info(f"Data written to {output_path}")
        except CLECompatibilityError as e:
            logger.info(f"Angr Error {filename} {e}")



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import uuid
from dataclasses import dataclass, field
from typing import List, Optional

from ._file import File
from ._provenance import HardwareProvenance

# pylint: disable=too-many-instance-attributes


@dataclass
class Hardware:
    UUID: str = field(default_factory=lambda: str(uuid.uuid4()))
    name: Optional[str] = None
    captureTime: Optional[int] = None
    officialName: Optional[str] = None
    countryOfOrigin: Optional[List[str]] = None
    countryOfOriginSource: Optional[str] = None
    quantity: Optional[int] = None
    description: Optional[str] = None
    vendor: Optional[List[str]] = None
    identifiers: Optional[List[str]] = None
    hardwareType: Optional[List[str]] = None
    comments: Optional[str] = None
    metadata: Optional[List[object]] = None
    supplementaryFiles: Optional[List[File]] = None
    packageType: Optional[str] = None
    boardLocation: Optional[List[str]] = None
    provenance: Optional[List[HardwareProvenance]] = None
    recordedInstitution: Optional[str] = None



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import os
import pathlib
import sys
import zlib

import pytest

from surfactant.filetypeid.id_magic import identify_file_type

# Files are relative to tests/data/
_file_to_file_type = {
    "a_out_files/big_m68020.aout": "A.OUT big",
    "a_out_files/big_netbsd_i386.aout": "A.OUT big",
    "a_out_files/big_netbsd_sparc.aout": "A.OUT big",
    "a_out_files/little_386.aout": "A.OUT little",
    "a_out_files/little_unknown.aout": "A.OUT little",
    "coff_files/intel_80386_coff": "COFF",
    "ELF_shared_obj_test_no1/bin/hello_world": "ELF",
    "ELF_shared_obj_test_no1/lib/libtestlib.so": "ELF",
    "java_class_no1/HelloWorld.class": "JAVACLASS",
    "mach_o_dylib_test_no1/bin/hello_world": "MACHO64",
    "mach_o_dylib_test_no1/lib/libtestlib.dylib": "MACHO64",
    "msitest_no1/test.msi": "OLE",
    "NET_app_config_test_no1/ConsoleApp2.exe": "PE",
    "NET_app_config_test_no1/bin/Debug/net6.0/hello.dll": "PE",
    "Windows_dll_test_no1/hello_world.exe": "PE",
    "Windows_dll_test_no1/testlib.dll": "PE",
}


def test_magic_id():
    base_path = pathlib.Path(__file__).parent.absolute()
    data_dir = os.path.join(base_path, "..", "data")
    for file_name, file_type in _file_to_file_type.items():
        assert identify_file_type(os.path.join(data_dir, file_name)) == file_type


def test_zlib_basic(tmp_path):
    for compress_level in range(10):
        write_to = tmp_path / f"basic_{compress_level}.zlib"
        write_to.write_bytes(zlib.compress(b"hello", level=compress_level))
        assert identify_file_type(write_to) == "ZLIB"


@pytest.mark.skipif(
    sys.version_info < (3, 11), reason="zlib.compress wbits only available from Python 3.11+"
)
def test_zlib_window(tmp_path):
    for compress_level in range(10):
        for window_size in range(9, 16):
            write_to = tmp_path / f"window_{compress_level}_{window_size}.zlib"
            write_to.write_bytes(zlib.compress(b"hello", level=compress_level, wbits=window_size))
            assert identify_file_type(write_to) == "ZLIB"



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import os
import pathlib
import queue
import re
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import click
from loguru import logger

from surfactant import ContextEntry
from surfactant.cmd.internal.generate_utils import SpecimenConfigParamType
from surfactant.configmanager import ConfigManager
from surfactant.fileinfo import sha256sum
from surfactant.plugin.manager import call_init_hooks, find_io_plugin, get_plugin_manager
from surfactant.relationships import parse_relationships
from surfactant.sbomtypes import SBOM, Software


# Converts from a true path to an install path
def real_path_to_install_path(root_path: str, install_path: str, filepath: str) -> str:
    # appending a "/" to root_path can cause problems if it is "" or already ends with "/"
    if root_path != "" and not root_path.endswith("/"):
        return re.sub("^" + root_path + "/", install_path, filepath)
    return re.sub("^" + root_path, install_path, filepath)


def get_software_entry(
    context,
    pluginmanager,
    parent_sbom: SBOM,
    filepath,
    *,  # arguments past this point are keyword-only
    filetype=None,
    container_uuid=None,
    root_path=None,
    install_path=None,
    user_institution_name="",
    omit_unrecognized_types=False,
) -> Tuple[Software, List[Software]]:
    sw_entry = Software.create_software_from_file(filepath)
    if root_path is not None and install_path is not None:
        sw_entry.installPath = [real_path_to_install_path(root_path, install_path, filepath)]
    if root_path is not None and container_uuid is not None:
        # make sure there is a "/" separating container uuid and the filepath
        if root_path != "" and not root_path.endswith("/"):
            sw_entry.containerPath = [re.sub("^" + root_path, container_uuid, filepath)]
        else:
            sw_entry.containerPath = [re.sub("^" + root_path, container_uuid + "/", filepath)]
    sw_entry.recordedInstitution = user_institution_name
    sw_children: List[Software] = []

    # for unsupported file types, details are just empty; this is the case for archive files (e.g. zip, tar, iso)
    # as well as intel hex or motorola s-rec files
    extracted_info_results: List[object] = pluginmanager.hook.extract_file_info(
        sbom=parent_sbom,
        software=sw_entry,
        filename=filepath,
        filetype=filetype,
        context=context,
        children=sw_children,
        omit_unrecognized_types=omit_unrecognized_types,
    )
    # add metadata extracted from the file, and set SBOM fields if metadata has relevant info
    for file_details in extracted_info_results:
        # None as details doesn't add any useful info...
        if file_details is None:
            continue

        # ensure metadata exists for the software entry
        if sw_entry.metadata is None:
            sw_entry.metadata = []
        sw_entry.metadata.append(file_details)

        # before checking for keys, make sure the file details object is a dictionary
        if not isinstance(file_details, Dict):
            continue

        # common case is Windows PE file has these details under FileInfo, otherwise fallback default value is fine
        if "FileInfo" in file_details:
            fi = file_details["FileInfo"]
            if "ProductName" in fi:
                sw_entry.name = fi["ProductName"]
            if "FileVersion" in fi:
                sw_entry.version = fi["FileVersion"]
            if "CompanyName" in fi:
                sw_entry.vendor = [fi["CompanyName"]]
            if "FileDescription" in fi:
                sw_entry.description = fi["FileDescription"]
            if "Comments" in fi:
                sw_entry.comments = fi["Comments"]

        # less common: OLE file metadata that might be relevant
        if filetype == "OLE":
            logger.trace("-----------OLE--------------")
            if "subject" in file_details["ole"]:
                sw_entry.name = file_details["ole"]["subject"]
            if "revision_number" in file_details["ole"]:
                sw_entry.version = file_details["ole"]["revision_number"]
            if "author" in file_details["ole"]:
                # ensure the vendor list has been created
                if sw_entry.vendor is None:
                    sw_entry.vendor = []
                sw_entry.vendor.append(file_details["ole"]["author"])
            if "comments" in file_details["ole"]:
                sw_entry.comments = file_details["ole"]["comments"]
    return (sw_entry, sw_children)


def print_output_formats(ctx, _, value):
    if not value or ctx.resilient_parsing:
        return
    pm = get_plugin_manager()
    for plugin in pm.get_plugins():
        if hasattr(plugin, "write_sbom"):
            if hasattr(plugin, "short_name"):
                print(plugin.short_name())
            else:
                print(pm.get_canonical_name(plugin))
    ctx.exit()


def print_input_formats(ctx, _, value):
    if not value or ctx.resilient_parsing:
        return
    pm = get_plugin_manager()
    for plugin in pm.get_plugins():
        if hasattr(plugin, "read_sbom"):
            if hasattr(plugin, "short_name"):
                print(plugin.short_name())
            else:
                print(pm.get_canonical_name(plugin))
    ctx.exit()


def determine_install_prefix(
    entry: Optional[ContextEntry] = None,
    extract_path: Optional[Union[str, pathlib.Path]] = None,
    skip_extract_path: bool = False,
) -> Optional[str]:
    """Determine the install prefix based on what is provided in the context entry, and the extract path for the file.

    Args:
        entry (Optional[ContextEntry]): The context entry to check for an install prefix.
        extract_path (Optional[str|pathlib.Path]): The extract path for the file to use as a potential fallback.
        skip_extract_path (bool): Whether the extract_path should be skipped if the entry does not specify an installPrefix.

    Returns:
            Optional[str]: The install prefix to use, or 'NoneType' if an install path shouldn't be listed.
    """
    install_prefix = None
    if entry and (entry.installPrefix or entry.installPrefix == ""):
        install_prefix = entry.installPrefix
    elif not skip_extract_path and extract_path is not None:
        # pathlib doesn't include the trailing slash
        epath = pathlib.Path(extract_path)
        if epath.is_file():
            install_prefix = epath.parent.as_posix() if len(epath.parts) > 1 else ""
        else:
            install_prefix = epath.as_posix()
        # add a trailing slash after last directory name
        if install_prefix != "" and not install_prefix.endswith("/"):
            install_prefix += "/"
    return install_prefix


def get_default_from_config(option: str, fallback: Optional[Any] = None) -> Any:
    """Retrive a core config option for use as default argument value.

    Args:
        option (str): The core config option to get.
        fallback (Optional[Any]): The fallback value if the option is not found.

    Returns:
            Any: The configuration value or 'NoneType' if the key doesn't exist.
    """
    config_manager = ConfigManager()
    return config_manager.get("core", option, fallback=fallback)


@click.command("generate")
@click.argument(
    "specimen_config",
    envvar="SPECIMEN_CONFIG",
    type=SpecimenConfigParamType(),
    required=True,
)
@click.argument("sbom_outfile", envvar="SBOM_OUTPUT", type=click.File("w"), required=True)
@click.argument("input_sbom", type=click.File("r"), required=False)
@click.option(
    "--skip_gather",
    is_flag=True,
    default=False,
    required=False,
    help="Skip gathering information on files and adding software entries",
)
@click.option(
    "--skip_relationships",
    is_flag=True,
    default=False,
    required=False,
    help="Skip adding relationships based on Linux/Windows/etc metadata",
)
@click.option(
    "--skip_install_path",
    is_flag=True,
    default=False,
    required=False,
    help="Skip including install path information if not given by configuration",
)
@click.option(
    "--recorded_institution",
    is_flag=False,
    default=get_default_from_config("recorded_institution"),
    help="Name of user's institution",
)
@click.option(
    "--output_format",
    is_flag=False,
    default=get_default_from_config("output_format", fallback="surfactant.output.cytrics_writer"),
    help="SBOM output format, see --list-output-formats for list of options; default is CyTRICS",
)
@click.option(
    "--list_output_formats",
    is_flag=True,
    callback=print_output_formats,
    expose_value=False,
    is_eager=True,
    help="List supported output formats",
)
@click.option(
    "--input_format",
    is_flag=False,
    default="surfactant.input_readers.cytrics_reader",
    help="Input SBOM format, see --list-input-formats for list of options; default is CyTRICS",
)
@click.option(
    "--list_input_formats",
    is_flag=True,
    callback=print_input_formats,
    expose_value=False,
    is_eager=True,
    help="List supported input formats",
)
@click.option(
    "--omit_unrecognized_types",
    is_flag=True,
    default=get_default_from_config("omit_unrecognized_types", fallback=False),
    required=False,
    help="Omit files with unrecognized types from the generated SBOM.",
)
# Disable positional argument linter check -- could make keyword-only, but then defaults need to be set
# pylint: disable-next=too-many-positional-arguments
def sbom(
    specimen_config: list,
    sbom_outfile: click.File,
    input_sbom: click.File,
    skip_gather: bool,
    skip_relationships: bool,
    skip_install_path: bool,
    recorded_institution: str,
    output_format: str,
    input_format: str,
    omit_unrecognized_types: bool,
):
    """Generate a sbom configured in CONFIG_FILE and output to SBOM_OUTPUT.

    An optional INPUT_SBOM can be supplied to use as a base for subsequent operations.
    """

    pm = get_plugin_manager()
    call_init_hooks(
        pm, hook_filter=["identify_file_type", "extract_file_info"], command_name="generate"
    )
    output_writer = find_io_plugin(pm, output_format, "write_sbom")
    input_reader = find_io_plugin(pm, input_format, "read_sbom")

    context: queue.Queue[ContextEntry] = queue.Queue()

    for cfg_entry in specimen_config:
        context.put(ContextEntry(**cfg_entry))

    # define the new_sbom variable type
    new_sbom: SBOM
    if not input_sbom:
        new_sbom = SBOM()
    else:
        new_sbom = input_reader.read_sbom(input_sbom)

    # gather metadata for files and add/augment software entries in the sbom
    if not skip_gather:
        # List of directory symlinks; 2-sized tuples with (source, dest)
        dir_symlinks: List[Tuple[str, str]] = []
        # List of file install path symlinks; keys are SHA256 hashes, values are source paths
        file_symlinks: Dict[str, List[str]] = {}
        # List of filename symlinks; keys are SHA256 hashes, values are file names
        filename_symlinks: Dict[str, List[str]] = {}
        while not context.empty():
            entry: ContextEntry = context.get()
            if entry.archive:
                logger.info("Processing parent container " + str(entry.archive))
                # TODO: if the parent archive has an info extractor that does unpacking interally, should the children be added to the SBOM?
                # current thoughts are (Syft) doesn't provide hash information for a proper SBOM software entry, so exclude these
                # extractor plugins meant to unpack files could be okay when used on an "archive", but then extractPaths should be empty
                parent_entry, _ = get_software_entry(
                    context,
                    pm,
                    new_sbom,
                    entry.archive,
                    user_institution_name=recorded_institution,
                )
                archive_entry = new_sbom.find_software(parent_entry.sha256)
                if (
                    archive_entry
                    and parent_entry
                    and Software.check_for_hash_collision(archive_entry, parent_entry)
                ):
                    logger.warning(
                        f"Hash collision between {archive_entry.name} and {parent_entry.name}; unexpected results may occur"
                    )
                if archive_entry:
                    parent_entry = archive_entry
                else:
                    new_sbom.add_software(parent_entry)
                parent_uuid = parent_entry.UUID
            else:
                parent_entry = None
                parent_uuid = None

            # If an installPrefix was given, clean it up some
            if entry.installPrefix:
                if not entry.installPrefix.endswith(("/", "\\")):
                    # Make sure the installPrefix given ends with a "/" (or Windows backslash path, but users should avoid those)
                    logger.warning(
                        "Fixing installPrefix in config file entry (include the trailing /)"
                    )
                    entry.installPrefix += "/"
                if "\\" in entry.installPrefix:
                    # Using an install prefix with backslashes can result in a gradual reduction of the number of backslashes... and weirdness
                    # Ideally even on a Windows "/" should be preferred instead in file paths, but "\" can be a valid character in Linux folder names
                    logger.warning(
                        "Fixing installPrefix with Windows-style backslash path separator in config file (ideally use / as path separator instead of \\, even for Windows"
                    )
                    entry.installPrefix = entry.installPrefix.replace("\\", "\\\\")

            for epath_str in entry.extractPaths:
                # convert to pathlib.Path, ensures trailing "/" won't be present and some more consistent path formatting
                epath = pathlib.Path(epath_str)
                install_prefix = determine_install_prefix(
                    entry, epath, skip_extract_path=skip_install_path
                )
                logger.trace("Extracted Path: " + epath.as_posix())

                # variable used to track software entries to add to the SBOM
                entries: List[Software]

                # handle individual file case, since os.walk doesn't
                if epath.is_file():
                    entries = []
                    filepath = epath.as_posix()
                    # breakpoint()
                    try:
                        sw_parent, sw_children = get_software_entry(
                            context,
                            pm,
                            new_sbom,
                            filepath,
                            filetype=pm.hook.identify_file_type(filepath=filepath),
                            root_path=epath.parent.as_posix() if len(epath.parts) > 1 else "",
                            container_uuid=parent_uuid,
                            install_path=install_prefix,
                            user_institution_name=recorded_institution,
                        )
                    except Exception as e:
                        raise RuntimeError(f"Unable to process: {filepath}") from e
                    entries.append(sw_parent)
                    entries.extend(sw_children if sw_children else [])
                    new_sbom.add_software_entries(entries, parent_entry=parent_entry)
                    # epath was a file, no need to walk the directory tree
                    continue

                # epath is a directory, walk it
                for cdir, dirs, files in os.walk(epath):
                    logger.info("Processing " + str(cdir))

                    if entry.installPrefix:
                        for dir_ in dirs:
                            full_path = os.path.join(cdir, dir_)
                            if os.path.islink(full_path):
                                dest = resolve_link(
                                    full_path, cdir, epath.as_posix(), entry.installPrefix
                                )
                                if dest is not None:
                                    install_source = real_path_to_install_path(
                                        epath.as_posix(), entry.installPrefix, full_path
                                    )
                                    install_dest = real_path_to_install_path(
                                        epath.as_posix(), entry.installPrefix, dest
                                    )
                                    dir_symlinks.append((install_source, install_dest))

                    entries = []
                    for file in files:
                        # os.path.join will insert an OS specific separator between cdir and f
                        # need to make sure that separator is a / and not a \ on windows
                        filepath = pathlib.Path(cdir, file).as_posix()
                        # TODO: add CI tests for generating SBOMs in scenarios with symlinks... (and just generally more CI tests overall...)
                        # Record symlink details but don't run info extractors on them
                        if os.path.islink(filepath):
                            # NOTE: resolve_link function could print warning if symlink goes outside of extract path dir
                            true_filepath = resolve_link(
                                filepath, cdir, epath.as_posix(), entry.installPrefix
                            )
                            # Dead/infinite links will error so skip them
                            if true_filepath is None:
                                continue
                            # Compute sha256 hash of the file; skip if the file pointed by the symlink can't be opened
                            try:
                                true_file_sha256 = sha256sum(true_filepath)
                            except FileNotFoundError:
                                logger.warning(
                                    f"Unable to open symlink {filepath} pointing to {true_filepath}"
                                )
                                continue
                            # Record the symlink name to be added as a file name
                            # Dead links would appear as a file, so need to check the true path to see
                            # if the thing pointed to is a file or a directory
                            if os.path.isfile(true_filepath):
                                if true_file_sha256 and true_file_sha256 not in filename_symlinks:
                                    filename_symlinks[true_file_sha256] = []
                                symlink_base_name = pathlib.PurePath(filepath).name
                                if symlink_base_name not in filename_symlinks[true_file_sha256]:
                                    filename_symlinks[true_file_sha256].append(symlink_base_name)
                            # Record symlink install path if an install prefix is given
                            if entry.installPrefix:
                                install_filepath = real_path_to_install_path(
                                    epath.as_posix(), entry.installPrefix, filepath
                                )
                                install_dest = real_path_to_install_path(
                                    epath.as_posix(), entry.installPrefix, true_filepath
                                )
                                # A dead link shows as a file so need to test if it's a
                                # file or a directory once rebased
                                if os.path.isfile(true_filepath):
                                    if true_file_sha256 and true_file_sha256 not in file_symlinks:
                                        file_symlinks[true_file_sha256] = []
                                    file_symlinks[true_file_sha256].append(install_filepath)
                                else:
                                    dir_symlinks.append((install_filepath, install_dest))
                            # NOTE Two cases that don't get recorded (but maybe should?) are:
                            # 1. If the file pointed to is outside the extract paths, it won't
                            # appear in the SBOM at all -- is that desirable? If it were included,
                            # should the true path also be included as an install path?
                            # 2. Does a symlink "exist" inside an archive/installer, or only after
                            # unpacking/installation?
                            continue

                        if os.path.isfile(filepath):
                            if not entry.includeFileExts:
                                entry.includeFileExts = []
                            if not entry.excludeFileExts:
                                entry.excludeFileExts = []
                            if (
                                ftype := pm.hook.identify_file_type(filepath=filepath)
                                or (not (omit_unrecognized_types or entry.omitUnrecognizedTypes))
                                or (
                                    os.path.splitext(filepath)[1].lower()
                                    in [ext.lower() for ext in entry.includeFileExts]
                                )
                            ) and os.path.splitext(filepath)[1].lower() not in [
                                ext.lower() for ext in entry.excludeFileExts
                            ]:
                                try:
                                    sw_parent, sw_children = get_software_entry(
                                        context,
                                        pm,
                                        new_sbom,
                                        filepath,
                                        filetype=ftype,
                                        root_path=epath.as_posix(),
                                        container_uuid=parent_uuid,
                                        install_path=install_prefix,
                                        user_institution_name=recorded_institution,
                                        omit_unrecognized_types=omit_unrecognized_types
                                        or entry.omitUnrecognizedTypes,
                                    )
                                except Exception as e:
                                    raise RuntimeError(f"Unable to process: {filepath}") from e

                                entries.append(sw_parent)
                                entries.extend(sw_children if sw_children else [])
                    new_sbom.add_software_entries(entries, parent_entry=parent_entry)

        # Add symlinks to install paths and file names
        for software in new_sbom.software:
            # ensure fileName, installPath, and metadata lists for the software entry have been created
            # for a user supplied input SBOM, there are no guarantees
            if software.fileName is None:
                software.fileName = []
            if software.installPath is None:
                software.installPath = []
            if software.metadata is None:
                software.metadata = []
            if software.sha256 in filename_symlinks:
                filename_symlinks_added = []
                for filename in filename_symlinks[software.sha256]:
                    if filename not in software.fileName:
                        software.fileName.append(filename)
                        filename_symlinks_added.append(filename)
                if filename_symlinks_added:
                    # Store information on which file names are symlinks
                    software.metadata.append({"fileNameSymlinks": filename_symlinks_added})
            if software.sha256 in file_symlinks:
                symlinks_added = []
                for full_path in file_symlinks[software.sha256]:
                    if full_path not in software.installPath:
                        software.installPath.append(full_path)
                        symlinks_added.append(full_path)
                if symlinks_added:
                    # Store information on which install paths are symlinks
                    software.metadata.append({"installPathSymlinks": symlinks_added})

        # Add directory symlink destinations to extract/install paths
        for software in new_sbom.software:
            # NOTE: this probably doesn't actually add any containerPath symlinks
            for paths in (software.containerPath, software.installPath):
                if paths is None:
                    continue
                paths_to_add = []
                for path in paths:
                    for link_source, link_dest in dir_symlinks:
                        if path.startswith(link_dest):
                            # Replace the matching start with the symlink instead
                            # We can't use os.path.join here because we end up with absolute paths after
                            # removing the common start.
                            paths_to_add.append(path.replace(link_dest, link_source, 1))
                if paths_to_add:
                    found_md_installpathsymlinks = False
                    # make sure software.metadata list has been initialized
                    if software.metadata is None:
                        software.metadata = []
                    if isinstance(software.metadata, Iterable):
                        for md in software.metadata:
                            if isinstance(md, Dict) and "installPathSymlinks" in md:
                                found_md_installpathsymlinks = True
                                md["installPathSymlinks"] += paths_to_add
                    if not found_md_installpathsymlinks:
                        software.metadata.append({"installPathSymlinks": paths_to_add})
                    paths += paths_to_add
    else:
        logger.info("Skipping gathering file metadata and adding software entries")

    # add "Uses" relationships based on gathered metadata for software entries
    if not skip_relationships:
        parse_relationships(pm, new_sbom)
    else:
        logger.info("Skipping relationships based on imports metadata")

    # TODO should contents from different containers go in different SBOM files, so new portions can be added bit-by-bit with a final merge?
    output_writer.write_sbom(new_sbom, sbom_outfile)


def resolve_link(
    path: str, cur_dir: str, extract_dir: str, install_prefix: Optional[str] = None
) -> Union[str, None]:
    assert cur_dir.startswith(extract_dir)
    # Links seen before
    seen_paths = set()
    # os.readlink() resolves one step of a symlink
    current_path = path
    steps = 0
    while os.path.islink(current_path):
        # If we've already seen this then we're in an infinite loop
        if current_path in seen_paths:
            logger.warning(f"Resolving symlink {path} encountered infinite loop at {current_path}")
            return None
        seen_paths.add(current_path)
        dest = os.readlink(current_path)
        # Convert relative paths to absolute local paths
        if not pathlib.Path(dest).is_absolute():
            common_path = os.path.commonpath([cur_dir, extract_dir])
            local_path = os.path.join("/", cur_dir[len(common_path) :])
            dest = os.path.join(local_path, dest)
        # Convert to a canonical form to eliminate .. to prevent reading above extract_dir
        # NOTE: should consider detecting reading above extract_dir and warn the user about incomplete file system structure issues
        dest = os.path.normpath(dest)
        if install_prefix and dest.startswith(install_prefix):
            dest = dest[len(install_prefix) :]
        # We need to get a non-absolute path so os.path.join works as we want
        if pathlib.Path(dest).is_absolute():
            # TODO: Windows support, but how???
            dest = dest[1:]
        # Rebase to get the true location
        current_path = os.path.join(extract_dir, dest)
        cur_dir = os.path.dirname(current_path)
    if not os.path.exists(current_path):
        logger.warning(f"Resolved symlink {path} to a path that doesn't exist {current_path}")
        return None
    return os.path.normpath(current_path)



from typing import Dict, List, Optional

import surfactant.plugin
from surfactant.sbomtypes import SBOM, Relationship, Software


def has_required_fields(metadata) -> bool:
    return "javaClasses" in metadata


class _ExportDict:
    created = False
    supplied_by: Dict[str, str] = {}

    @classmethod
    def create_export_dict(cls, sbom: SBOM):
        if cls.created:
            return
        for software_entry in sbom.software:
            if software_entry.metadata:
                for metadata in software_entry.metadata:
                    if isinstance(metadata, Dict) and "javaClasses" in metadata:
                        for class_info in metadata["javaClasses"].values():
                            for export in class_info["javaExports"]:
                                cls.supplied_by[export] = software_entry.UUID
        cls.created = True

    @classmethod
    def get_supplier(cls, import_name: str) -> Optional[str]:
        if import_name in cls.supplied_by:
            return cls.supplied_by[import_name]
        return None


@surfactant.plugin.hookimpl
def establish_relationships(
    sbom: SBOM, software: Software, metadata
) -> Optional[List[Relationship]]:
    if not has_required_fields(metadata):
        return None
    _ExportDict.create_export_dict(sbom)
    relationships = []
    dependant_uuid = software.UUID
    for class_info in metadata["javaClasses"].values():
        for import_ in class_info["javaImports"]:
            if supplier_uuid := _ExportDict.get_supplier(import_):
                if supplier_uuid != dependant_uuid:
                    rel = Relationship(dependant_uuid, supplier_uuid, "Uses")
                    if rel not in relationships:
                        relationships.append(rel)
    return relationships



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from surfactant.plugin.manager import get_plugin_manager
from surfactant.sbomtypes import SBOM, Relationship, Software

sbom = SBOM(
    software=[
        Software(
            UUID="abc",
            fileName=["helics_broker"],
            installPath=["/usr/local/bin/helics_broker"],
            metadata=[
                {
                    "elfDependencies": ["libhelicscpp-apps.so"],
                    "elfRpath": [],
                    "elfRunpath": ["$ORIGIN:$ORIGIN/../lib:/usr/lib/x86_64-linux-gnu"],
                    "elfDynamicFlags": [{"DF_ORIGIN": False}],
                    "elfDynamicFlags1": [{"DF_1_ORIGIN": False, "DF_1_NODEFLIB": False}],
                }
            ],
        ),
        Software(
            UUID="xyz",
            fileName=["libhelicscpp-apps.so"],
            installPath=["/usr/local/lib/libhelicscpp-apps.so"],
            metadata=[
                {
                    "elfDependencies": ["libzmq.so"],
                    "elfRpath": [],
                    "elfRunpath": ["$ORIGIN:$ORIGIN/../lib:/usr/lib/x86_64-linux-gnu"],
                    "elfDynamicFlags": [],
                    "elfDynamicFlags1": [],
                }
            ],
        ),
        Software(
            UUID="def",
            fileName=["libzmq.so"],
            installPath=["/lib/libzmq.so", "/customlib/abspath/libzmq.so"],
            metadata=[
                {
                    "elfDependencies": [],
                    "elfRpath": [],
                    "elfRunpath": [],
                    "elfDynamicFlags": [],
                    "elfDynamicFlags1": [],
                }
            ],
        ),
        Software(
            UUID="hij",
            fileName=["libcomm.so"],
            installPath=["/customlib/relpath/misc/libcomm.so"],
            metadata=[
                {
                    "elfDependencies": ["/customlib/abspath/libzmq.so"],
                    "elfRpath": [],
                    "elfRunpath": [],
                    "elfDynamicFlags": [],
                    "elfDynamicFlags1": [],
                }
            ],
        ),
        Software(
            UUID="klm",
            fileName=["libcomm-cpp.so"],
            installPath=["/customlib/relpath/libcomm-cpp.so"],
            metadata=[
                {
                    "elfDependencies": ["misc/libcomm.so"],
                    "elfRpath": [],
                    "elfRunpath": [],
                    "elfDynamicFlags": [],
                    "elfDynamicFlags1": [],
                }
            ],
        ),
    ],
    relationships=[],
)


def test_relative_paths():
    elfPlugin = get_plugin_manager().get_plugin("surfactant.relationships.elf_relationship")
    sw = sbom.software[4]
    md = sw.metadata[0]
    # located in /customlib/relpath/misc, dependency specified as being under misc/ relative path
    assert elfPlugin.establish_relationships(sbom, sw, md) == [Relationship("klm", "hij", "Uses")]


def test_absolute_paths():
    elfPlugin = get_plugin_manager().get_plugin("surfactant.relationships.elf_relationship")
    sw = sbom.software[3]
    md = sw.metadata[0]
    # located in /customlib/abspath
    assert elfPlugin.establish_relationships(sbom, sw, md) == [Relationship("hij", "def", "Uses")]


def test_default_system_paths():
    elfPlugin = get_plugin_manager().get_plugin("surfactant.relationships.elf_relationship")
    sw = sbom.software[1]
    md = sw.metadata[0]
    # located in /lib
    assert elfPlugin.establish_relationships(sbom, sw, md) == [Relationship("xyz", "def", "Uses")]


def test_dst_expansion():
    elfPlugin = get_plugin_manager().get_plugin("surfactant.relationships.elf_relationship")
    sw = sbom.software[0]
    md = sw.metadata[0]
    # uses origin expansion
    assert elfPlugin.establish_relationships(sbom, sw, md) == [Relationship("abc", "xyz", "Uses")]



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import pathlib
from collections.abc import Iterable
from typing import List, Optional

import surfactant.plugin
from surfactant.sbomtypes import SBOM, Relationship, Software

from ._internal.windows_utils import find_installed_software


def has_required_fields(metadata) -> bool:
    return "peImport" in metadata or "peBoundImport" in metadata or "peDelayImport" in metadata


@surfactant.plugin.hookimpl
def establish_relationships(
    sbom: SBOM, software: Software, metadata
) -> Optional[List[Relationship]]:
    if not has_required_fields(metadata):
        return None

    relationships = []
    if "peImport" in metadata:
        # NOTE: UWP apps have their own search order for libraries; they use a .appx or .msix file extension and appear to be zip files, so our SBOM probably doesn't even include them
        relationships.extend(get_windows_pe_dependencies(sbom, software, metadata["peImport"]))
    if "peBoundImport" in metadata:
        relationships.extend(get_windows_pe_dependencies(sbom, software, metadata["peBoundImport"]))
    if "peDelayImport" in metadata:
        relationships.extend(get_windows_pe_dependencies(sbom, software, metadata["peDelayImport"]))
    return relationships


def get_windows_pe_dependencies(sbom: SBOM, sw: Software, peImports) -> List[Relationship]:
    relationships: List[Relationship] = []
    # No installPath is probably temporary files/installer
    # TODO maybe resolve dependencies using relative locations in containerPath, for files originating from the same container UUID?
    if sw.installPath is None:
        return relationships

    # https://learn.microsoft.com/en-us/windows/win32/dlls/dynamic-link-library-search-order
    # Desktop Applications (we can only check a subset of these without much more info gathering, disassembly + full filesystem + environment details)
    # 1. Specifying full path, using DLL redirection, or using a manifest
    # - https://learn.microsoft.com/en-us/windows/win32/dlls/dynamic-link-library-redirection
    # - DLL redirection summary: redirection file with name_of_exe.local (contents are ignored) makes a check for mydll.dll happen in the application directory first, regardless of what the full path specified for LoadLibrary or LoadLibraryEx is (if no dll found in local directory, uses the typical search order)
    # - manifest files cause any .local files to be ignored (also, enabling DLL redirection may require setting DevOverrideEnable registry key)
    # 2. If DLL with same module name is loaded in memory, no search will happen. If DLL is in KnownDLLs registry key, it uses the system copy of the DLL instead of searching.
    # 3. If LOAD_LIBRARY_SEARCH flags are set for LoadLibraryEx, it will search dir LOAD_LIBRARY_SEARCH_DLL_LOAD_DIR, LOAD_LIBRARY_SEARCH_APPLICATION_DIR, paths explicitly added by AddDllDirectory (LOAD_LIBRARY_SEARCH_USER_DIRS) or the SetDllDirectory (multiple paths added have unspecified search order), then system directory (LOAD_LIBRARY_SEARCH_SYSTEM32)
    # 4. Look in dir the app was loaded from (or specified by absolute path lpFileName if LoadLibraryEx is called with LOAD_WITH_ALTERED_SEARCH_PATH)
    # 5. If SetDllDirectory function called with lpPathName: the directory specified
    # 6. If SafeDllSearchMode is disabled: the current directory
    # 7. Look in the system directory (GetSystemDirectory to get the path)
    # 8. The 16-bit system directory (no function to get this directory; %windir%\SYSTEM on 32-bit systems, not supported on 64-bit systems)
    # 9. Windows system directory (GetWindowsDirectory to get this path)
    # 10. If SafeDllSearchMode is enabled (default): the current directory
    # 11. Directories listed in PATH environment variable (per-application path in App Paths registry key is not used for searching)

    # In addition, Windows 10 + 11 add a feature called API sets: https://learn.microsoft.com/en-us/windows/win32/apiindex/windows-apisets
    # these use special dll names that aren't actually a physical file on disk

    # Of those steps, without gathering much more information that is likely not available or manual/dynamic analysis, we can do:
    # 4. Look for DLL in the directory the application was loaded from
    dependent_uuid = sw.UUID
    for fname in peImports:
        probedirs = []
        if isinstance(sw.installPath, Iterable):
            for ipath in sw.installPath:
                probedirs.append(pathlib.PureWindowsPath(ipath).parent.as_posix())
        # likely just one found, unless sw entry has the same file installed to multiple places
        for e in find_installed_software(sbom, probedirs, fname):
            dependency_uuid = e.UUID
            relationships.append(Relationship(dependent_uuid, dependency_uuid, "Uses"))
        # logging DLLs not found would be nice, but is excessively noisy due being almost exclusively system DLLs

    return relationships



import json
from pathlib import Path

from surfactant.cmd.generate import sbom

testing_data = Path(Path(__file__).parent.parent, "data")


def test_generate_no_install_prefix(tmp_path):
    extract_path = Path(testing_data, "Windows_dll_test_no1").as_posix()
    config_data = f'[{{"extractPaths": ["{extract_path}"]}}]'
    config_path = str(Path(tmp_path, "config.json"))
    output_path = str(Path(tmp_path, "out.json"))

    with open(config_path, "w") as f:
        f.write(config_data)

    # the click.testing module would be better here but it doesn't allow for files to be generated
    # pylint: disable=no-value-for-parameter
    sbom([config_path, output_path], standalone_mode=False)
    # pylint: enable

    with open(output_path) as f:
        generated_sbom = json.load(f)

    assert len(generated_sbom["software"]) == 2

    expected_software_names = {"hello_world.exe", "testlib.dll"}
    actual_software_names = {software["fileName"][0] for software in generated_sbom["software"]}
    assert expected_software_names == actual_software_names

    expected_install_paths = {
        "hello_world.exe": extract_path + "/hello_world.exe",
        "testlib.dll": extract_path + "/testlib.dll",
    }
    for software in generated_sbom["software"]:
        assert software["installPath"][0] == expected_install_paths[software["fileName"][0]]

    uuids = {software["fileName"][0]: software["UUID"] for software in generated_sbom["software"]}
    assert len(generated_sbom["relationships"]) == 1
    assert generated_sbom["relationships"][0] == {
        "xUUID": uuids["hello_world.exe"],
        "yUUID": uuids["testlib.dll"],
        "relationship": "Uses",
    }


def test_generate_with_install_prefix(tmp_path):
    extract_path = Path(testing_data, "Windows_dll_test_no1").as_posix()
    config_data = f'[{{"extractPaths": ["{extract_path}"], "installPrefix": "test_prefix/"}}]'
    config_path = str(Path(tmp_path, "config.json"))
    output_path = str(Path(tmp_path, "out.json"))

    with open(config_path, "w") as f:
        f.write(config_data)

    # the click.testing module would be better here but it doesn't allow for files to be generated
    # pylint: disable=no-value-for-parameter
    sbom([config_path, output_path], standalone_mode=False)
    # pylint: enable

    with open(output_path) as f:
        generated_sbom = json.load(f)

    assert len(generated_sbom["software"]) == 2

    expected_software_names = {"hello_world.exe", "testlib.dll"}
    actual_software_names = {software["fileName"][0] for software in generated_sbom["software"]}
    assert expected_software_names == actual_software_names

    expected_install_paths = {
        "hello_world.exe": "test_prefix/hello_world.exe",
        "testlib.dll": "test_prefix/testlib.dll",
    }
    for software in generated_sbom["software"]:
        assert software["installPath"][0] == expected_install_paths[software["fileName"][0]]

    uuids = {software["fileName"][0]: software["UUID"] for software in generated_sbom["software"]}
    assert len(generated_sbom["relationships"]) == 1
    assert generated_sbom["relationships"][0] == {
        "xUUID": uuids["hello_world.exe"],
        "yUUID": uuids["testlib.dll"],
        "relationship": "Uses",
    }


def test_generate_with_skip_install_path(tmp_path):
    extract_path = Path(testing_data, "Windows_dll_test_no1").as_posix()
    config_data = f'[{{"extractPaths": ["{extract_path}"]}}]'
    config_path = str(Path(tmp_path, "config.json"))
    output_path = str(Path(tmp_path, "out.json"))

    with open(config_path, "w") as f:
        f.write(config_data)

    # the click.testing module would be better here but it doesn't allow for files to be generated
    # pylint: disable=no-value-for-parameter
    sbom(["--skip_install_path", config_path, output_path], standalone_mode=False)
    # pylint: enable

    with open(output_path) as f:
        generated_sbom = json.load(f)

    assert len(generated_sbom["software"]) == 2

    expected_software_names = {"hello_world.exe", "testlib.dll"}
    actual_software_names = {software["fileName"][0] for software in generated_sbom["software"]}
    assert expected_software_names == actual_software_names

    for software in generated_sbom["software"]:
        assert software["installPath"] == []

    assert len(generated_sbom["relationships"]) == 0



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from surfactant.plugin.manager import get_plugin_manager
from surfactant.sbomtypes import SBOM, Relationship, Software

sbom = SBOM(
    software=[
        Software(
            UUID="supplier",
            fileName=["supplier"],
            installPath=["supplier"],
            metadata=[{"javaClasses": {"dummy": {"javaExports": ["someFunc():void"]}}}],
        ),
        Software(
            UUID="consumer",
            fileName=["consumer"],
            installPath=["consumer"],
            metadata=[
                {
                    "javaClasses": {
                        "dummy": {
                            "javaExports": [],
                            "javaImports": ["someFunc():void"],
                        },
                    },
                },
            ],
        ),
    ],
    relationships=[],
)


def test_java_relationship():
    javaPlugin = get_plugin_manager().get_plugin("surfactant.relationships.java_relationship")
    sw = sbom.software[1]
    md = sw.metadata[0]
    assert javaPlugin.establish_relationships(sbom, sw, md) == [
        Relationship("consumer", "supplier", "Uses")
    ]



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from queue import Queue
from typing import List, Optional

from pluggy import HookspecMarker

from surfactant import ContextEntry
from surfactant.sbomtypes import SBOM, Relationship, Software

hookspec = HookspecMarker("surfactant")


@hookspec(firstresult=True)
def identify_file_type(filepath: str) -> Optional[str]:
    """Determine the type of file located at filepath, and return a string identifying the type
    that will be passed to file extraction plugins. Return `None` to indicate that the type was
    unable to be determined.

    Args:
        filepath (str): The path to the file to determine the type of.

    Returns:
        Optional[str]: A string identifying the type of file, or None if the file type could not be recognized.
    """


@hookspec
# pylint: disable-next=too-many-positional-arguments
def extract_file_info(
    sbom: SBOM,
    software: Software,
    filename: str,
    filetype: str,
    context: "Queue[ContextEntry]",
    children: List[Software],
    omit_unrecognized_types: bool,
) -> object:
    """Extracts information from the given file to add to the given software entry. Return an
    object to be included as part of the metadata field, and potentially used as part of
    selecting default values for other Software entry fields. Returning `None` will not add
    anything to the Software entry metadata.

    Args:
        sbom (SBOM): The SBOM that the software entry is part of. Can be used to add observations or analysis data.
        software (Software): The software entry the gathered information will be added to.
        filename (str): The full path to the file to extract information from.
        filetype (str): File type information based on magic bytes.
        context (Queue[ContextEntry]): Modifiable queue of entries from input config file. Existing plugins should still work without adding this parameter.
        children (List[Software]): List of additional software entries to include in the SBOM. Plugins can add additional entries, though if the plugin extracts files to a temporary directory, the context argument should be used to have Surfactant process the files instead.
        omit_unrecognized_types (bool): Whether files with types that are not recognized by Surfactant should be left out of the SBOM. When a plugin is adding additional context entries to the queue, it should typically default to propagating this value to the new context entries that it creates.

    Returns:
        object: An object to be added to the metadata field for the software entry. May be `None` to add no metadata.
    """


@hookspec
def establish_relationships(
    sbom: SBOM, software: Software, metadata
) -> Optional[List[Relationship]]:
    """Called to add relationships to an SBOM after information has been gathered.

    The function will be called once for every metadata object in every software
    entry in the SBOM. Realistically, this means a plugin should not be trying to
    establish relationships for the entire SBOM before returning, just for the
    software/metadata object that has been passed to it.

    Returns a list of relationships to be added to the SBOM.

    Args:
        sbom (SBOM): The SBOM object that the Software is part of.
        software (Software): The Software entry that the metadata object is from.
        metadata: The metadata object to establish relationships based on.

    Returns:
        Optional[List[Relationship]]: A list of relationships to add to the SBOM.
    """


@hookspec
def write_sbom(sbom: SBOM, outfile) -> None:
    """Writes the contents of the SBOM to the given output file.

    Args:
        sbom (SBOM): The SBOM to write to the output file.
        outfile: The output file handle to write the SBOM to.
    """


@hookspec
# type: ignore[empty-body]
def read_sbom(infile) -> SBOM:
    """Reads the contents of the input SBOM from the given input SBOM file.

    Returns a SBOM object containing the input SBOM, which can be added to.

    Args:
        infile: The input file handle to read the SBOM from.
    """


@hookspec
def short_name() -> Optional[str]:
    """A short name to register the hook as.

    Returns:
        Optional[str]: The name to register the hook with.
    """


@hookspec
def update_db() -> Optional[str]:
    """Updates the database for the plugin.

    This hook should be implemented by plugins that require a database update.
    The implementation should perform the necessary update operations.

    Returns:
        Optional[str]: A message indicating the result of the update operation, or None if no update was needed.
    """


@hookspec
def init_hook(command_name: Optional[str] = None) -> None:
    """Initialization hook for plugins.

    This hook is called to perform any necessary initialization for the plugin,
    such as loading databases or setting up resources.

    Args:
        command_name (Optional[str]): The name of the command invoking the initialization,
                                      which can be used to conditionally initialize based on the context.
    """



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import pathlib
import platform
import time
import uuid
from collections.abc import Iterable
from dataclasses import dataclass, field, fields
from typing import Any, List, Optional

from dataclasses_json import dataclass_json

from surfactant.fileinfo import calc_file_hashes, get_file_info

from ._file import File
from ._provenance import SoftwareComponentProvenance, SoftwareProvenance

# pylint: disable=too-many-instance-attributes


@dataclass_json
@dataclass
class SoftwareComponent:
    name: str
    captureTime: Optional[int] = None
    version: Optional[str] = None
    vendor: Optional[List[str]] = None
    description: Optional[str] = None
    comments: Optional[str] = None
    metadata: Optional[List[object]] = None
    supplementaryFiles: Optional[List[File]] = None
    provenance: Optional[List[SoftwareComponentProvenance]] = None
    recordedInstitution: Optional[str] = None


@dataclass_json
@dataclass
class Software:
    UUID: str = field(default_factory=lambda: str(uuid.uuid4()))
    name: Optional[str] = None
    size: Optional[int] = None
    fileName: Optional[List[str]] = None
    installPath: Optional[List[str]] = None
    containerPath: Optional[List[str]] = None
    captureTime: Optional[int] = None
    version: Optional[str] = None
    vendor: Optional[List[str]] = None
    description: Optional[str] = None
    sha1: Optional[str] = None
    sha256: Optional[str] = None
    md5: Optional[str] = None
    relationshipAssertion: Optional[str] = (
        None  # enum: Unknown, Root, Partial, Known; default=Unknown
    )
    comments: Optional[str] = None
    metadata: Optional[List[object]] = None
    supplementaryFiles: Optional[List[File]] = None
    provenance: Optional[List[SoftwareProvenance]] = None
    recordedInstitution: Optional[str] = None
    components: Optional[List[SoftwareComponent]] = None

    def _update_field(self, field_name: str, value: Any):
        if value not in ["", " ", None]:
            setattr(self, field_name, value)

    @staticmethod
    def create_software_from_file(filepath) -> Software:
        file_hashes = calc_file_hashes(filepath)
        stat_file_info = get_file_info(filepath)

        # add basic file info, and information on what collected the information listed for the file to aid later processing
        collection_info = {
            "collectedBy": "Surfactant",
            "collectionPlatform": platform.platform(),
            "fileInfo": {
                "mode": stat_file_info["filemode"],
                "hidden": stat_file_info["filehidden"],
            },
        }

        sw = Software(
            sha1=file_hashes["sha1"],
            sha256=file_hashes["sha256"],
            md5=file_hashes["md5"],
            fileName=[pathlib.Path(filepath).name],
            installPath=[],
            containerPath=[],
            size=stat_file_info["size"],
            captureTime=int(time.time()),
            version="",
            vendor=[],
            description="",
            relationshipAssertion="Unknown",
            comments="",
            metadata=[collection_info],
            supplementaryFiles=[],
            provenance=None,
            components=[],
        )
        return sw

    # TODO: figure out how to handle merging an SBOM with manual additions
    def merge(self, sw: Software):
        # hashes should be confirmed to match before calling this function
        # check to make sure entry isn't an exact duplicate
        if sw and self != sw:
            # leave UUID and captureTime the same
            single_value_fields = [
                "name",
                "comments",
                "version",
                "description",
                "relationshipAssertion",
                "recordedInstitution",
            ]
            array_fields = [
                "containerPath",
                "fileName",
                "installPath",
                "vendor",
                "provenance",
                "metadata",
                "supplementaryFiles",
                "components",
            ]
            for fld in fields(self):
                if fld.name in single_value_fields:
                    current_value = getattr(self, fld.name)
                    new_value = getattr(sw, fld.name)
                    if current_value != new_value:
                        self._update_field(fld.name, new_value)
                # for lists, append new values that we don't currently have
                if fld.name in array_fields:
                    current_arr = getattr(self, fld.name)
                    new_arr = getattr(sw, fld.name)
                    # if the multi-value fields differ, the one with new values *must* be Iterable
                    if current_arr != new_arr and isinstance(new_arr, Iterable):
                        # if our field is not iterable, initialize it as a new list
                        if current_arr is None:
                            setattr(self, fld.name, [])
                            current_arr = getattr(self, fld.name)
                        for new_value in new_arr:
                            # special case, UUID in containerPaths need updating to match our UUID
                            if fld.name == "containerPath":
                                if new_value.startswith(sw.UUID):
                                    new_value = new_value.replace(sw.UUID, self.UUID)
                            if new_value not in current_arr:
                                current_arr.append(new_value)

        return self.UUID, sw.UUID

    @staticmethod
    def check_for_hash_collision(soft1: Optional[Software], soft2: Optional[Software]) -> bool:
        if not soft1 or not soft2:
            return False
        # A hash collision occurs if one or more but less than all hashes match or
        # any hash matches but the filesize is different
        collision = False
        if soft1.sha256 == soft2.sha256 or soft1.sha1 == soft2.sha1 or soft1.md5 == soft2.md5:
            # Hashes can be None; make sure they aren't before checking for inequality
            if soft1.sha256 and soft2.sha256 and soft1.sha256 != soft2.sha256:
                collision = True
            elif soft1.sha1 and soft2.sha1 and soft1.sha1 != soft2.sha1:
                collision = True
            elif soft1.md5 and soft2.md5 and soft1.md5 != soft2.md5:
                collision = True
            elif soft1.size != soft2.size:
                collision = True
        return collision



# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import hashlib
from typing import Any, Dict, Optional

import requests
import tomlkit
from loguru import logger
from requests.exceptions import RequestException


def download_database(url: str) -> Optional[str]:
    """
    Downloads the content of a database from the given URL.

    Args:
        url (str): The URL of the database to download.
    Returns:
        Optional[str]: The content of the database as a string if the request is successful,
                       or None if an error occurs.
    """
    try:
        # Perform the HTTP GET request with a timeout
        response = requests.get(url, timeout=10)

        # Handle HTTP status codes
        if response.status_code == 200:
            logger.info(f"Request successful! URL: {url}")
            return response.text

        if response.status_code == 404:
            logger.error(f"Resource not found. URL: {url}")
        else:
            logger.warning(f"Unexpected status code {response.status_code} for URL: {url}")
    except RequestException as e:
        # Handle network-related errors
        logger.error(f"An error occurred while trying to fetch the URL: {url}. Error: {e}")

    # Return None in case of any failure
    return None


def calculate_hash(data: str) -> str:
    """
    Calculate the SHA-256 hash of the given data.

    Args:
        data (str): The input string to hash.

    Returns:
        str: The SHA-256 hash of the input string.
    """
    return hashlib.sha256(data.encode("utf-8")).hexdigest()


def _read_toml_file(file_path: str) -> Optional[Dict[str, Any]]:
    """
    Read and parse a TOML file.

    Args:
        file_path (str): The path to the TOML file.

    Returns:
        Optional[Dict[str, Any]]: The parsed TOML data, or None if the file does not exist.
    """
    try:
        with open(file_path, "r") as f:
            return tomlkit.load(f)
    except FileNotFoundError:
        return None
    except tomlkit.exceptions.TOMLKitError as e:
        raise ValueError(f"Error parsing TOML file at {file_path}: {e}") from e


def _write_toml_file(file_path: str, data: Dict[str, Any]) -> None:
    """
    Write data to a TOML file.

    Args:
        file_path (str): The path to the TOML file.
        data (Dict[str, Any]): The data to write to the file.
    """
    with open(file_path, "w") as f:
        tomlkit.dump(data, f)


def load_hash_and_timestamp(
    hash_file_path: str, pattern_key: str, pattern_file: str
) -> Optional[Dict[str, str]]:
    """
    Load the hash and timestamp for a specific pattern from the specified TOML file.

    Args:
        hash_file_path (str): The path to the TOML file.
        pattern_key (str): The key identifying the pattern group.
        pattern_file (str): The key identifying the specific pattern file.

    Returns:
        Optional[Dict[str, str]]: The hash and timestamp data, or None if not found.
    """
    hash_data = _read_toml_file(hash_file_path)
    if hash_data is None:
        return None

    # Access the specific structure using the provided keys
    return hash_data.get(pattern_key, {}).get(pattern_file)


def save_hash_and_timestamp(hash_file_path: str, pattern_info: Dict[str, str]) -> None:
    """
    Save the hash and timestamp for a specific pattern to the specified TOML file.

    Args:
        hash_file_path (str): The path to the TOML file.
        pattern_info (Dict[str, str]): A dictionary containing the following keys:
            - "pattern_key": The key identifying the pattern group.
            - "pattern_file": The key identifying the specific pattern file.
            - "source": The source of the pattern.
            - "hash_value": The hash value of the pattern.
            - "timestamp": The timestamp of the pattern.

    Raises:
        ValueError: If required keys are missing from `pattern_info`.
    """
    required_keys = {"pattern_key", "pattern_file", "source", "hash_value", "timestamp"}
    if not required_keys.issubset(pattern_info):
        raise ValueError(f"pattern_info must contain the keys: {required_keys}")

    hash_data = _read_toml_file(hash_file_path) or {}

    # Define the new data structure
    new_data = {
        pattern_info["pattern_key"]: {
            pattern_info["pattern_file"]: {
                "source": pattern_info["source"],
                "hash": pattern_info["hash_value"],
                "timestamp": pattern_info["timestamp"],
            }
        }
    }

    # Update the existing data with the new data
    if pattern_info["pattern_key"] in hash_data:
        hash_data[pattern_info["pattern_key"]].update(new_data[pattern_info["pattern_key"]])
    else:
        hash_data.update(new_data)

    # Write the updated data back to the TOML file
    _write_toml_file(hash_file_path, hash_data)



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

# TODO: Ensure all the relevant information available is extracted from here
# https://lief.re/doc/stable/api/python/macho.html

from sys import modules
from typing import Any, Dict

from loguru import logger

try:
    import lief
except ModuleNotFoundError:
    pass

import surfactant.plugin
from surfactant.configmanager import ConfigManager
from surfactant.infoextractors.__macho_cpuSubtypes import get_cpu_subtype_name, get_cpu_type_name
from surfactant.sbomtypes import SBOM, Software

__config_manager = ConfigManager()

__include_bindings_exports = __config_manager.get("macho", "include_bindings_exports", False)
__include_signature_content = __config_manager.get("macho", "include_signature_content", False)


def supports_file(filetype) -> bool:
    return filetype in ("MACHOFAT", "MACHOFAT64", "MACHO32", "MACHO64")


@surfactant.plugin.hookimpl
def extract_file_info(sbom: SBOM, software: Software, filename: str, filetype: str) -> object:
    if not supports_file(filetype):
        return None
    if "lief" not in modules:
        logger.warning("LIEF not installed - Skipped extracting Mach-O file info.")
        return None
    return extract_mach_o_info(filename)


def extract_mach_o_info(filename: str) -> object:
    try:
        binaries = lief.MachO.parse(filename)
    except OSError:
        return {}

    file_details: Dict[str, Any] = {"OS": "MacOS", "numBinaries": binaries.size, "binaries": []}

    # Iterate over all binaries in the FAT binary
    for binary in binaries:
        header = binary.header
        details = {
            "format": binary.format.__name__,
            "header": {
                "cpuType": get_cpu_type_name(header.cpu_type.value),
                "cpuTypeValue": header.cpu_type.value,
                "cpuSubtype": get_cpu_subtype_name(header.cpu_type.value, 20 + header.cpu_subtype),
                "cpuSubtypeValue": header.cpu_subtype,
                "fileType": header.file_type.__name__,
                "fileTypeValue": header.file_type.value,
                "flags": [flag.__name__ for flag in header.flags_list],
                "numCommands": header.nb_cmds,
            },
            "build": {},
            "signature": {},
            "dependencies": [],
            "rpaths": [],
            "dyld": {},
            "encryption": {},
        }

        # Extract info from build version
        if binary.has_build_version:
            build = binary.build_version
            details["build"] = {
                "platform": build.platform.__name__,
                "platformValue": build.platform.value,
                "minOSVersion": ".".join(map(str, build.minos)),
                "sdkVersion": ".".join(map(str, build.sdk)),
                "tools": [],
            }
            for tool in build.tools:
                details["build"]["tools"].append(
                    {"tool": tool.tool.__name__, "version": ".".join(map(str, tool.version))}
                )

        # Extract info from code signature
        if binary.has_code_signature or binary.has_code_signature_dir:
            if binary.has_code_signature:
                signature = binary.code_signature
                signature_type = "Default"
            else:
                signature = binary.code_signature_dir
                signature_type = "LC_DYLIB_CODE_SIGN_DRS"

            details["signature"] = {
                "offset": signature.data_offset,
                "size": signature.data_size,
                "type": signature_type,
            }
            if __include_signature_content:
                details["signature"]["content"] = signature.content

        # Extract library dependencies
        for library in binary.libraries:
            details["dependencies"].append(
                {
                    "name": library.name,
                    "currentVersion": ".".join(map(str, library.current_version)),
                    "compatibilityVersion": ".".join(map(str, library.compatibility_version)),
                }
            )

        # rpath info
        if binary.has_rpath:
            for rpath in binary.rpaths:
                details["rpaths"].append(rpath.path)

        # dyld info
        if binary.has_dylinker:
            details["dyld"]["linker"] = binary.dylinker.name
        if binary.has_dyld_environment:
            details["dyld"]["environment"] = binary.dyld_environment.__name__
            details["dyld"]["environmentValue"] = binary.dyld_environment.value
        # https://github.com/qyang-nj/llios/blob/main/dynamic_linking/chained_fixups.md
        if __include_bindings_exports:
            if binary.has_dyld_info:
                details["dyld"]["info"] = {
                    "bindings": [],
                    "exports": [],
                }
                bindings = get_bindings(binary.dyld_info.bindings)
                details["dyld"]["info"]["bindings"] = bindings
                for export in binary.dyld_info.exports:
                    details["dyld"]["info"]["exports"].append(
                        {"address": export.address, "kind": export.kind.__name__}
                    )
            elif binary.has_dyld_exports_trie and binary.has_dyld_chained_fixups:
                details["dyld"]["exports"] = []
                for export in binary.dyld_exports_trie.exports:
                    details["dyld"]["exports"].append(
                        {"address": export.address, "kind": export.kind.__name__}
                    )
                fixups = binary.dyld_chained_fixups
                details["dyld"]["chainedFixups"] = {"chainedStartsInSegment": [], "bindings": []}
                for chainedStarts in fixups.chained_starts_in_segments:
                    details["dyld"]["chainedFixups"]["chainedStartsInSegment"].append(
                        {
                            "maxValidPointer": chainedStarts.max_valid_pointer,
                            "pageSize": chainedStarts.page_size,
                            "pointerFormat": chainedStarts.pointer_format.__name__,
                            "pointerFormatValue": chainedStarts.pointer_format.value,
                            "segment": chainedStarts.segment.name,
                        }
                    )
                bindings = get_bindings(fixups.bindings)
                details["dyld"]["chainedFixups"]["bindings"] = bindings

        # encryption info
        if binary.has_encryption_info:
            encryption = binary.encryption_info
            details["encryption"] = {
                "system": encryption.crypt_id,
                "offset": encryption.crypt_offset,
                "size": encryption.crypt_size,
            }
        file_details["binaries"].append(details)

    return file_details


def get_bindings(bindings):
    bindings_info = []
    for binding in bindings:
        # ChainedFixups and dyld info store bindings in the same format in LIEF
        info = {
            "addend": binding.addend,
            "address": binding.address,
            "isWeakImport": binding.weak_import,
        }
        if binding.has_library:
            info["library"] = binding.library.name
        if binding.has_segment:
            info["segment"] = binding.segment.name
        bindings_info.append(info)
    return bindings_info



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
from pathlib import Path

from checksec.__main__ import checksec_file
from checksec.elf import ELFChecksecData
from checksec.pe import PEChecksecData

import surfactant.plugin
from surfactant.sbomtypes import SBOM, Software


@surfactant.plugin.hookimpl
def extract_file_info(sbom: SBOM, software: Software, filename: str, filetype: str) -> object:
    if filetype not in ["ELF", "PE"]:
        return None
    checksec_data = checksec_file(Path(filename))
    data = {}
    if isinstance(checksec_data, ELFChecksecData):
        data["checksecDataType"] = "ELF"
        data["checksec"] = {
            "relro": checksec_data.relro.name,
            "canary": checksec_data.canary,
            "nx": checksec_data.nx,
            "pie": checksec_data.pie.name,
            "rpath": checksec_data.rpath,
            "runpath": checksec_data.runpath,
            "symbols": checksec_data.symbols,
            "fortify_source": checksec_data.fortify_source,
            "fortified": checksec_data.fortified,
            "fortify-able": checksec_data.fortifiable,
            "fortify_score": checksec_data.fortify_score,
        }
    elif isinstance(checksec_data, PEChecksecData):
        data["checksecDataType"] = "PE"
        data["checksec"] = {
            "nx": checksec_data.nx,
            "canary": checksec_data.canary,
            "aslr": checksec_data.aslr,
            "dynamic_base": checksec_data.dynamic_base,
            "high_entropy_va": checksec_data.high_entropy_va,
            "isolation": checksec_data.isolation,
            "seh": checksec_data.seh,
            "safe_seh": checksec_data.safe_seh,
            "authenticode": checksec_data.authenticode,
            "guard_cf": checksec_data.guard_cf,
            "force_integrity": checksec_data.force_integrity,
        }
    return data



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import surfactant.plugin
from surfactant.sbomtypes import SBOM, Software


def supports_file(filetype: str) -> bool:
    return filetype == "COFF"


@surfactant.plugin.hookimpl
def extract_file_info(sbom: SBOM, software: Software, filename: str, filetype: str) -> object:
    if not supports_file(filetype):
        return None
    return extract_coff_out_info(filetype, filename)


# Machine types:
# https://github.com/file/file/blob/master/magic/Magdir/coff
COFF_MAGIC_TARGET_NAME = {
    0x014C: "Intel 80386",
    0x0500: "Hitachi SH big-endian",
    0x0550: "Hitachi SH little-endian",
    0x0200: "Intel ia64",
    0x8664: "Intel amd64",
    0xAA64: "Aarch64",
    0x01C0: "ARM",
    0xA641: "ARM64EC",
    0x01C2: "ARM Thumb",
    0x01C4: "ARMv7 Thumb",
}


def extract_coff_out_info(filetype: str, filename: str) -> object:
    try:
        with open(filename, "rb") as f:
            magic_bytes = f.read(4)
            magic_int = int.from_bytes(magic_bytes, byteorder="little", signed=False)
            if magic_int in COFF_MAGIC_TARGET_NAME:
                return {"coffMachineType": COFF_MAGIC_TARGET_NAME[magic_int]}
            return None
    except FileNotFoundError:
        return None



# Copyright 2024 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import json
import os
from typing import Optional

import click
import textual.app
import textual.containers
import textual.css.query
import textual.events
import textual.screen
import textual.types
import textual.widgets
import textual.widgets.button


class YesNoScreen(textual.screen.Screen[bool]):
    """Screen that presents a yes/no question"""

    def __init__(self, question: str) -> None:
        self.question = question
        super().__init__()

    def compose(self) -> textual.app.ComposeResult:
        yield textual.widgets.Label(self.question)
        yield textual.widgets.Button("Yes", id="yes", variant="success")
        yield textual.widgets.Button("No", id="no")

    @textual.on(textual.widgets.Button.Pressed, "#yes")
    def handle_yes(self) -> None:
        self.dismiss(True)

    @textual.on(textual.widgets.Button.Pressed, "#no")
    def handle_no(self) -> None:
        self.dismiss(False)


class SelectFileButtons(textual.widgets.Static):
    # pylint: disable=too-few-public-methods
    def __init__(self, allow_folder_selection: bool):
        super().__init__()
        self.allow_folder_selection = allow_folder_selection

    def compose(self) -> textual.app.ComposeResult:
        yield textual.widgets.Button("Go up a directory", id="up_dir")
        if self.allow_folder_selection:
            yield textual.widgets.Button("Select directory", id="select_dir")


class SelectFile(textual.screen.ModalScreen[Optional[textual.widgets.DirectoryTree.FileSelected]]):
    """Pop-up to select a file"""

    def __init__(self, allow_folder_selection: bool):
        super().__init__()
        self.allow_folder_selection = allow_folder_selection
        self.dir_selected = "./"

    def compose(self) -> textual.app.ComposeResult:
        yield textual.widgets.DirectoryTree("./", id="file_dir")
        yield SelectFileButtons(self.allow_folder_selection)

    @textual.on(textual.widgets.Button.Pressed, "#up_dir")
    def handle_up_dir(self):
        tree = self.get_child_by_id("file_dir", textual.widgets.DirectoryTree)
        tree.path = tree.path.parent.resolve()
        tree.reload()

    @textual.on(textual.widgets.Button.Pressed, "#select_dir")
    def handle_select_dir(self):
        self.dismiss(self.dir_selected)

    def on_directory_tree_directory_selected(
        self, path: textual.widgets.DirectoryTree.DirectorySelected
    ) -> None:
        self.dir_selected = path.path.as_posix()

    def on_directory_tree_file_selected(
        self, path: textual.widgets.DirectoryTree.FileSelected
    ) -> None:
        self.dismiss(path.path.as_posix())

    def on_key(self, event: textual.events.Key):
        if event.key == "escape":
            self.dismiss(None)


class ExtractPathContent(textual.widgets.Static):
    def __init__(self, parent):
        super().__init__()
        self.par = parent

    def compose(self) -> textual.app.ComposeResult:
        if len(self.par.path) > 0:
            yield textual.widgets.Label(f"{self.par.path}", id="path")
        else:
            yield textual.widgets.Label("(Click to set extract path)", id="path")

    def on_click(self):
        def set_path(path: Optional[textual.widgets.DirectoryTree.FileSelected]):
            if path:
                self.par.path = path
                self.get_child_by_id("path", textual.widgets.Label).update(f"Path: {self.par.path}")

        self.app.push_screen(SelectFile(True), set_path)


class ExtractPathSelector(textual.widgets.Static):
    def __init__(self, path, **kwargs):
        super().__init__(**kwargs)
        self.path = path
        self.alive = True

    def compose(self) -> textual.app.ComposeResult:
        yield textual.widgets.Button("-", id="delete_path")
        yield ExtractPathContent(self)

    @textual.on(textual.widgets.Button.Pressed, "#delete_path")
    def on_delete_path(self):
        self.remove()
        self.alive = False


class ExtractPathAdder(textual.widgets.Static):
    """A widget for adding extract path selectors"""

    def __init__(self, entry_id):
        super().__init__()
        self.paths = []
        self.entry_id = entry_id
        for path in self.app.config_data[entry_id]["extractPaths"]:
            self.paths.append(ExtractPathSelector(path))

    def compose(self) -> textual.app.ComposeResult:
        yield textual.widgets.Button("+", id=f"add_path_{self.entry_id}")

    def on_mount(self):
        entries = self.app.query_one("#config_entries", textual.containers.ScrollableContainer)
        for path in self.paths:
            if path.alive:
                entries.mount(path, before=f"#add_path_{self.entry_id}")

    def on_button_pressed(self, event: textual.widgets.Button.Pressed) -> None:
        if event.button.id == f"add_path_{self.entry_id}":
            entries = self.app.query_one("#config_entries", textual.containers.ScrollableContainer)
            self.paths.append(ExtractPathSelector(""))
            entries.mount(self.paths[-1], before=f"#add_path_{self.entry_id}")


class ArchiveEntry(textual.widgets.Static):
    def __init__(self, entry_id):
        super().__init__()
        self.archive_loc = self.app.config_data[entry_id]["archive"]

    def compose(self) -> textual.app.ComposeResult:
        if len(self.archive_loc) > 0:
            yield textual.widgets.Label(f"Archive: {self.archive_loc}", id="archive_label")
        else:
            yield textual.widgets.Label("Archive: [Click to select archive]", id="archive_label")

    def on_click(self):
        def set_archive(path: Optional[textual.widgets.DirectoryTree.FileSelected]):
            if path:
                self.archive_loc = path
                self.get_child_by_id("archive_label", textual.widgets.Label).update(
                    f"Archive: {path}"
                )

        self.app.push_screen(SelectFile(False), set_archive)


class InstallPrefix(textual.widgets.Static):
    def __init__(self, entry_id):
        super().__init__()
        self.install_prefix = self.app.config_data[entry_id]["installPrefix"]

    def compose(self) -> textual.app.ComposeResult:
        yield textual.widgets.Label("Install Prefix: ")
        yield textual.widgets.Input(self.install_prefix, id="prefix")

    @textual.on(textual.widgets.Input.Changed, "#prefix")
    def on_prefix_change(self):
        self.install_prefix = self.query_one("#prefix", textual.widgets.Input).value


class ConfigEntry(textual.widgets.Static):
    """A widget for holding a single entry within a config file"""

    def __init__(self, entry_id):
        super().__init__()
        self.entry_id = entry_id
        self.archive_entry = ArchiveEntry(entry_id)
        self.install_prefix = InstallPrefix(entry_id)
        self.extract_path_adder = ExtractPathAdder(entry_id)
        self.alive = True

    def compose(self) -> textual.app.ComposeResult:
        yield textual.widgets.Button("Delete this entry", id="delete_entry")
        yield self.archive_entry
        yield self.install_prefix
        yield textual.widgets.Label("Extract Paths:")
        yield self.extract_path_adder

    @textual.on(textual.widgets.Button.Pressed, "#delete_entry")
    def on_delete_entry(self):
        def delete_self(do_it: Optional[bool]):
            if do_it:
                self.remove()
                # entries = self.app.query_one(
                #     "#config_entries", textual.containers.ScrollableContainer
                # )
                # for child in entries.children:
                #     child.remove()
                # self.app.query_one(f"#view_entry_{self.entry_id}").remove()
                self.alive = False

        self.app.push_screen(
            YesNoScreen("Are you sure you want to delete this entry?"), delete_self
        )


def add_entry(app: textual.app.App) -> ConfigEntry:
    if app.entry_id not in app.config_data:
        app.config_data[app.entry_id] = {
            "archive": "",
            "extractPaths": [],
            "installPrefix": "",
        }
    app.config_entries[app.entry_id] = ConfigEntry(app.entry_id)
    app.config_entries[app.entry_id].border_title = str(app.entry_id)
    app.entry_id += 1
    return app.config_entries[app.entry_id - 1]


class AddEntry(textual.widgets.Static):
    """A widget for adding another config entry"""

    def compose(self) -> textual.app.ComposeResult:
        yield textual.widgets.Button("+", id="add_config_entry")

    @textual.on(textual.widgets.Button.Pressed, "#add_config_entry")
    def on_add_config_entry(self):
        self.app.query_one("#config_entries").mount(add_entry(self.app), before="#add_config_entry")


class ConfigTUI(textual.app.App):
    """An app for creating and modifying Surfactant config files"""

    BINDINGS = [
        ("d", "toggle_dark", "Toggle dark mode"),
        ("s", "save_sbom", "Save SBOM"),
        ("q", "quit", "Quit"),
    ]
    TITLE = "Surfactant Config TUI"
    CSS_PATH = "../web-files/config_tui.tcss"

    def __init__(self, config_path: str):
        super().__init__()
        if os.path.exists(config_path):
            with open(config_path, "r") as config_file:
                self.config_json = json.load(config_file)
        else:
            self.config_json = []

        self.config_path = config_path
        self.sub_title = config_path
        self.config_entries = {}
        self.config_data = {}
        self.entry_id = 1

    def compose(self) -> textual.app.ComposeResult:
        yield textual.widgets.Header()
        yield textual.widgets.Footer()
        yield textual.containers.ScrollableContainer(id="config_entries")

    def on_mount(self):
        for i, entry in enumerate(self.config_json):
            i += 1
            self.config_data[i] = {"archive": "", "extractPaths": [], "installPrefix": ""}
            if "archive" in entry:
                self.config_data[i]["archive"] = entry["archive"]
            if "extractPaths" in entry:
                self.config_data[i]["extractPaths"] = entry["extractPaths"]
            if "installPrefix" in entry:
                self.config_data[i]["installPrefix"] = entry["installPrefix"]
            add_entry(self)
        for entry in self.config_entries.values():
            self.query_one("#config_entries").mount(entry)
        self.query_one("#config_entries").mount(AddEntry())

    def action_toggle_dark(self) -> None:
        """A binding for toggling dark mode"""
        # pylint: disable=attribute-defined-outside-init
        self.theme = "textual-dark" if self.theme == "textual-light" else "textual-light"

    def action_save_sbom(self) -> None:
        """Saves the current SBOM to file"""
        to_save = []
        for entry in self.config_entries.values():
            if entry.alive:
                to_save.append({})
                write_to = to_save[-1]
                # Load data, ignoring it if empty
                archive = entry.archive_entry.archive_loc
                if len(archive) > 0:
                    write_to["archive"] = archive
                install_prefix = entry.install_prefix.install_prefix
                if len(install_prefix) > 0:
                    write_to["installPrefix"] = install_prefix
                # Always write extract paths
                write_to["extractPaths"] = []
                extract_path_adder = entry.extract_path_adder
                for selector in extract_path_adder.paths:
                    if selector.alive:
                        write_to["extractPaths"].append(selector.path)

        with open(self.config_path, "w") as f:
            f.write(json.dumps(to_save, indent=2))

    def action_quit(self) -> None:
        self.action_save_sbom()
        self.app.exit()


@click.command("config_tui")
@click.argument("config_file", type=str, required=True)
def config_tui(config_file):
    """Create a configuration input file with a TUI"""
    app = ConfigTUI(config_file)
    app.run()
